import asyncio
import json
import logging
import os
import shutil
import string
import tempfile
import uuid
import zipfile
from asyncio import Future, Queue
from contextlib import asynccontextmanager
from datetime import timedelta, datetime, timezone
from functools import partial
from io import BytesIO
from pathlib import Path
from random import SystemRandom
from typing import (
    AsyncGenerator,
    Any,
    Optional,
    Sequence,
    Union,
    List,
    Dict,
    AsyncIterator,
    Tuple,
    Callable,
    Awaitable,
    Iterable,
)
from urllib.parse import urlencode, urlparse, parse_qs, urlunparse

import aiohttp_jinja2
import jinja2
import prometheus_client
import yaml
from aiohttp import (
    web,
    MultipartWriter,
    AsyncIterablePayload,
    BufferedReaderPayload,
    MultipartReader,
    ClientSession,
    TCPConnector,
)
from aiohttp.abc import AbstractStreamWriter
from aiohttp.hdrs import METH_ANY
from aiohttp.web import Request, StreamResponse, WebSocketResponse
from aiohttp.web_exceptions import HTTPNotFound, HTTPNoContent, HTTPOk, HTTPNotAcceptable, HTTPSeeOther
from aiohttp.web_fileresponse import FileResponse
from aiohttp.web_response import json_response
from aiohttp_swagger3 import SwaggerFile, SwaggerUiSettings
from aiostream import stream
from attrs import evolve
from dateutil import parser as date_parser
from networkx.readwrite import cytoscape_data
from resotoui import ui_path

from resotocore.analytics import AnalyticsEvent
from resotocore.cli.command import alias_names
from resotocore.dependencies import Dependencies
from resotocore.cli.model import (
    ParsedCommandLine,
    CLIContext,
    CLICommand,
    InternalPart,
    WorkerCustomCommand,
    AliasTemplate,
    InfraAppAlias,
    FilePath,
)
from resotocore.config import ConfigValidation, ConfigEntity
from resotocore.console_renderer import ConsoleColorSystem, ConsoleRenderer
from resotocore.db.graphdb import GraphDB, HistoryChange
from resotocore.db.model import QueryModel
from resotocore.error import NotFoundError
from resotocore.ids import TaskId, ConfigId, NodeId, SubscriberId, WorkerId, GraphName, Email, Password
from resotocore.message_bus import Message, ActionDone, Action, ActionError, ActionInfo, ActionProgress
from resotocore.metrics import timed
from resotocore.model.graph_access import Section
from resotocore.model.json_schema import json_schema
from resotocore.model.model import Kind, Model
from resotocore.model.typed_model import to_json, from_js, to_js_str, to_js
from resotocore.service import Service
from resotocore.task.model import Subscription
from resotocore.types import Json, JsonElement
from resotocore.util import uuid_str, force_gen, rnd_str, if_set, duration, utc_str, parse_utc, async_noop, utc
from resotocore.web.auth import raw_jwt_from_auth_message, AuthorizedUser, AuthHandler
from resotocore.web.content_renderer import result_binary_gen, single_result
from resotocore.web.directives import (
    metrics_handler,
    error_handler,
    on_response_prepare,
    cors_handler,
    enable_compression,
    default_middleware,
)
from resotocore.web.tsdb import tsdb
from resotocore.worker_task_queue import (
    WorkerTaskDescription,
    WorkerTask,
    WorkerTaskResult,
    WorkerTaskInProgress,
)
from resotolib.asynchronous.web.ws_handler import accept_websocket, clean_ws_handler
from resotolib.jwt import encode_jwt
from resotolib.x509 import cert_to_bytes

log = logging.getLogger(__name__)


def section_of(request: Request) -> Optional[str]:
    section = request.match_info.get("section", request.query.get("section"))
    if section and section != "/" and section not in Section.content:
        raise AttributeError(f"Given section does not exist: {section}")
    return section


# No Authorization required for following paths
AlwaysAllowed = {
    "/",
    "/.well-known/.*",
    "/api-doc.*",
    "/authenticate",
    "/ca/cert",
    "/create-first-user",
    "/login",
    "/metrics",
    "/static/.*",
    "/system/.*",
    "/ui.*",
    "/notebook/.*",
    "/debug/.*",
}
# Authorization is not required, but implemented as part of the request handler
DeferredCheck = {"/events", "/work/queue"}


class Api(Service):
    def __init__(self, deps: Dependencies):
        self.deps = deps
        self.auth_handler = AuthHandler(
            deps.db_access.system_data_db, deps.config, deps.cert_handler, AlwaysAllowed | DeferredCheck
        )

        self.app = web.Application(
            client_max_size=self.deps.config.api.max_request_size or 1024**2,
            # note on order: the middleware is passed in the order provided.
            middlewares=[
                metrics_handler,
                self.auth_handler.middleware(),
                cors_handler,
                error_handler(deps.config, deps.event_sender),
                default_middleware(self),
            ],
        )
        self.app.on_response_prepare.append(on_response_prepare)
        self._session: Optional[ClientSession] = None
        self.in_shutdown = False
        self.websocket_handler: Dict[str, Tuple[Future[Any], WebSocketResponse]] = {}
        path_part = deps.config.api.web_path.strip().strip("/").strip()
        web_path = "" if path_part == "" else f"/{path_part}"
        self.__add_routes(web_path)
        aiohttp_jinja2.setup(
            self.app, loader=jinja2.FileSystemLoader(os.path.abspath(os.path.dirname(__file__) + "/../templates"))
        )

    @property
    def session(self) -> ClientSession:
        if self._session is None:
            # only keep connections alive for 15 seconds, cleanup closed transports
            connector = TCPConnector(keepalive_timeout=15.0, enable_cleanup_closed=True)
            self._session = ClientSession(connector=connector)
        return self._session

    def __add_routes(self, prefix: str) -> None:
        static_path = os.path.abspath(os.path.dirname(__file__) + "/../static")
        jupyterlite_path = Path(os.path.abspath(os.path.dirname(__file__) + "/../jupyterlite"))
        if not jupyterlite_path.exists():
            jupyterlite_path.mkdir(parents=True, exist_ok=True)
        self.app.add_routes(
            [
                # Model operations (backwards compatible)
                web.get(prefix + "/model", self.get_model),
                web.patch(prefix + "/model", self.update_model),
                web.get(prefix + "/model/uml", self.model_uml),
                # Graph based model operations
                web.get(prefix + "/graph/{graph_id}/model", self.get_model),
                web.patch(prefix + "/graph/{graph_id}/model", self.update_model),
                web.get(prefix + "/graph/{graph_id}/model/uml", self.model_uml),
                # CRUD Graph operations
                web.get(prefix + "/graph", self.list_graphs),
                web.get(prefix + "/graph/{graph_id}", self.get_node),
                web.post(prefix + "/graph/{graph_id}", self.create_graph),
                web.delete(prefix + "/graph/{graph_id}", self.wipe),
                # search the graph
                web.post(prefix + "/graph/{graph_id}/search/raw", self.raw),
                web.post(prefix + "/graph/{graph_id}/search/structure", self.query_structure),
                web.post(prefix + "/graph/{graph_id}/search/explain", self.explain),
                web.post(prefix + "/graph/{graph_id}/search/list", self.query_list),
                web.post(prefix + "/graph/{graph_id}/search/graph", self.query_graph_stream),
                web.post(prefix + "/graph/{graph_id}/search/aggregate", self.query_aggregation),
                web.post(prefix + "/graph/{graph_id}/search/history/list", self.query_history),
                web.post(prefix + "/graph/{graph_id}/search/history/aggregate", self.query_history),
                # maintain the graph
                web.patch(prefix + "/graph/{graph_id}/nodes", self.update_nodes),
                web.post(prefix + "/graph/{graph_id}/merge", self.merge_graph),
                web.post(prefix + "/graph/{graph_id}/batch/merge", self.update_merge_graph_batch),
                web.get(prefix + "/graph/{graph_id}/batch", self.list_batches),
                web.post(prefix + "/graph/{graph_id}/batch/{batch_id}", self.commit_batch),
                web.delete(prefix + "/graph/{graph_id}/batch/{batch_id}", self.abort_batch),
                # node specific actions
                web.post(prefix + "/graph/{graph_id}/node/{node_id}/under/{parent_node_id}", self.create_node),
                web.get(prefix + "/graph/{graph_id}/node/{node_id}", self.get_node),
                web.patch(prefix + "/graph/{graph_id}/node/{node_id}", self.update_node),
                web.delete(prefix + "/graph/{graph_id}/node/{node_id}", self.delete_node),
                web.patch(prefix + "/graph/{graph_id}/node/{node_id}/section/{section}", self.update_node),
                # Subscriptions
                web.get(prefix + "/subscribers", self.list_all_subscriptions),
                web.get(prefix + "/subscribers/for/{event_type}", self.list_subscription_for_event),
                # Subscription
                web.get(prefix + "/subscriber/{subscriber_id}", self.get_subscriber),
                web.put(prefix + "/subscriber/{subscriber_id}", self.update_subscriber),
                web.delete(prefix + "/subscriber/{subscriber_id}", self.delete_subscriber),
                web.post(prefix + "/subscriber/{subscriber_id}/{event_type}", self.add_subscription),
                web.delete(prefix + "/subscriber/{subscriber_id}/{event_type}", self.delete_subscription),
                web.get(prefix + "/subscriber/{subscriber_id}/handle", self.handle_subscribed),
                # report checks
                web.get(prefix + "/report/checks", self.inspection_checks),
                web.get(prefix + "/report/checks/graph/{graph_id}", self.perform_benchmark_on_checks),
                web.get(prefix + "/report/check/{check_id}/graph/{graph_id}", self.inspection_results),
                web.get(prefix + "/report/benchmark/{benchmark}/graph/{graph_id}", self.perform_benchmark),
                # CLI
                web.post(prefix + "/cli/evaluate", self.evaluate),
                web.post(prefix + "/cli/execute", self.execute),
                web.get(prefix + "/cli/info", self.cli_info),
                # Event operations
                web.get(prefix + "/events", self.handle_events),
                web.post(prefix + "/analytics", self.send_analytics_events),
                # Worker operations
                web.get(prefix + "/work/queue", self.handle_work_tasks),
                web.get(prefix + "/work/list", self.list_work),
                # Serve static filed
                web.get(prefix, self.forward("/ui/index.html")),
                web.static(prefix + "/static", static_path),
                web.get(prefix + "/notebook", self.forward("/notebook/index.html")),
                web.static(prefix + "/notebook", jupyterlite_path),
                # metrics
                web.get(prefix + "/metrics", self.metrics),
                # config operations
                web.get(prefix + "/configs", self.list_configs),
                web.patch(prefix + "/config/{config_id:[^{}]+}", self.patch_config),
                web.delete(prefix + "/config/{config_id:[^{}]+}", self.delete_config),
                # config model operations
                web.get(prefix + "/configs/validation", self.list_config_models),
                web.get(prefix + "/configs/model", self.get_configs_model),
                web.patch(prefix + "/configs/model", self.update_configs_model),
                web.put(prefix + "/config_validation/{config_id:[^{}]+}", self.put_config_validation),
                web.get(prefix + "/config_validation/{config_id:[^{}]+}", self.get_config_validation),
                web.put(prefix + "/config/{config_id:[^{}]+}/validation", self.put_config_validation),
                web.get(prefix + "/config/{config_id:[^{}]+}/validation", self.get_config_validation),
                # config operations, moved here to avoid early matching
                web.put(prefix + "/config/{config_id:[^{}]+}", self.put_config),
                web.get(prefix + "/config/{config_id:[^{}]+}", self.get_config),
                # ca operations
                web.get(prefix + "/ca/cert", self.certificate),
                web.post(prefix + "/ca/sign", self.sign_certificate),
                # system operations
                web.get(prefix + "/system/ping", self.ping),
                web.get(prefix + "/system/ready", self.ready),
                # forwards
                web.get(prefix + "/tsdb", self.forward("/tsdb/")),
                web.get(prefix + "/ui", self.forward("/ui/index.html")),
                web.get(prefix + "/ui/", self.forward("/ui/index.html")),
                # auth operations
                web.get(prefix + "/.well-known/jwks.json", self.jwks),
                web.get(prefix + "/login", self.login_page),
                web.post(prefix + "/create-first-user", self.create_first_user),
                web.post(prefix + "/authenticate", self.authenticate),
                web.get(prefix + "/authorization/user", self.get_authorized_user),
                web.get(prefix + "/authorization/renew", self.renew_authorization),
                # tsdb operations
                web.route(METH_ANY, prefix + "/tsdb/{tail:.+}", tsdb(self)),
                web.static(f"{prefix}/ui/", ui_path),
            ]
        )
        if self.deps.config.runtime.debug:
            self.app.add_routes([web.get(prefix + "/debug/ui/{commit}/{path:.+}", self.serve_debug_ui)])
        SwaggerFile(
            self.app,
            spec_file=f"{static_path}/api-doc.yaml",
            swagger_ui_settings=SwaggerUiSettings(path=prefix + "/api-doc", layout="BaseLayout", docExpansion="none"),
        )

    async def start(self) -> None:
        await self.auth_handler.start()

    async def stop(self) -> None:
        await self.auth_handler.stop()
        if not self.in_shutdown:
            self.in_shutdown = True
            for ws_id in list(self.websocket_handler):
                await clean_ws_handler(ws_id, self.websocket_handler)
            if self.session:
                await self.session.close()

    @staticmethod
    async def login_with_redirect(request: Request) -> StreamResponse:
        params = request.query.copy()
        params["redirect"] = request.raw_path
        return web.HTTPSeeOther("/login?" + urlencode(params))

    @staticmethod
    def forward(to: str) -> Callable[[Request], Awaitable[StreamResponse]]:
        async def forward_to(request: Request) -> StreamResponse:
            goto = to + "?" + urlencode(request.query) if request.query else to
            return web.HTTPFound(goto)

        return forward_to

    @staticmethod
    async def ping(_: Request) -> StreamResponse:
        return web.HTTPOk(text="pong", content_type="text/plain")

    @staticmethod
    async def ready(_: Request) -> StreamResponse:
        return web.HTTPOk(text="ok")

    async def jwks(self, _: Request) -> StreamResponse:
        return web.json_response(self.auth_handler.signing_key_jwk)

    async def login_page(self, request: Request) -> StreamResponse:
        template = "login.html" if await self.deps.user_management.has_users() else "create_first_user.html"
        return aiohttp_jinja2.render_template(template, request, context=request.query)

    async def create_first_user(self, request: Request) -> StreamResponse:
        post_data = await request.post()
        errors = []
        try:
            email = Email(str(post_data.get("email", "")).strip())
            password = Password(str(post_data.get("password", "")))
            password_repeat = str(post_data.get("password_repeat", ""))
            company = str(post_data.get("company", "")).strip()
            fullname = str(post_data.get("fullname", "")).strip()
            if not email or email.startswith("@") or email.endswith("@") or email.count("@") != 1:
                errors.append("Invalid email address")
            if not password:
                errors.append("Password is required")
            if not company:
                errors.append("Company name is required")
            if not fullname:
                errors.append("Full name is required")
            if password != password_repeat:
                errors.append("Passwords do not match")
            if not errors:
                await self.deps.user_management.create_first_user(company, fullname, email, password)
                return await self.authenticate(request)
        except Exception as e:
            errors.append(str(e))
        error_string = ". ".join(errors)
        return aiohttp_jinja2.render_template(
            "create_first_user.html", request, context={**post_data, "error": error_string}
        )

    async def authenticate(self, request: Request) -> StreamResponse:
        post_data = await request.post()
        email = Email(str(post_data.get("email", "")))
        password = Password(str(post_data.get("password", "")))
        redirect = str(post_data.get("redirect", ""))
        if email and password and (user := await self.deps.user_management.login(email, password)):
            params: Dict[str, List[str]] = {}
            if self.deps.config.args.psk:
                code = await self.auth_handler.add_authorized_user(AuthorizedUser(email, user.roles, utc()))
                params["code"] = [code]
            if redirect:
                if params:
                    parsed = urlparse(redirect)
                    query_params = parse_qs(parsed.query)
                    query_params.update(params)
                    parsed = parsed._replace(query=urlencode(query_params, doseq=True))
                    redirect = urlunparse(parsed)
                response: StreamResponse = HTTPSeeOther(redirect)
            else:
                response = HTTPOk(text=urlencode(params, doseq=True))
            return response
        return aiohttp_jinja2.render_template(
            "login.html", request, context=dict(**post_data, error="Invalid username or password")
        )

    @staticmethod
    async def get_authorized_user(request: Request) -> StreamResponse:
        if jwt := request.get("jwt"):
            return web.json_response(jwt)
        else:
            return web.HTTPNoContent()

    async def renew_authorization(self, request: Request) -> StreamResponse:
        if jwt_raw := request.get("jwt"):
            exp = datetime.fromtimestamp(int(jwt_raw["exp"]), tz=timezone.utc)
            user = AuthorizedUser(jwt_raw["email"], set(jwt_raw["roles"].split(",")), exp)
            renewed, data = self.auth_handler.user_jwt(user)
            return web.json_response(data, headers={"Authorization": f"Bearer {renewed}"})
        else:
            return HTTPNoContent()  # no psk, no renewal

    async def list_configs(self, request: Request) -> StreamResponse:
        return await self.stream_response_from_gen(request, self.deps.config_handler.list_config_ids())

    async def get_config(self, request: Request) -> StreamResponse:
        config_id = ConfigId(request.match_info["config_id"])
        accept = request.headers.get("accept", "application/json")
        not_found = HTTPNotFound(text="No config with this id")
        if accept == "application/yaml":
            yml = await self.deps.config_handler.config_yaml(config_id)
            return web.Response(body=yml.encode("utf-8"), content_type="application/yaml") if yml else not_found
        else:

            def get_query_param_bool(name: str, default: bool) -> bool:
                return request.query.get(name, "true" if default else "false").lower() == "true"

            # do we want the config with overrides/env_vars applied in-place or in a separate object?
            separate_overrides = get_query_param_bool("separate_overrides", default=False)

            if separate_overrides:
                # if we want separate overrides, we don't apply overrides to the existing config
                # and don't substitute the env vars by default. E.g. UI asks us for the config
                apply_overrides = get_query_param_bool("apply_overrides", default=False)
                resolve_env_vars = get_query_param_bool("resolve_env_vars", default=False)
                # attach the "raw" config version that was stored in the database
                include_raw_config = get_query_param_bool("include_raw_config", default=False)
            else:
                # if we request a single object with overrides applied,
                # we apply overrides and resolve env vars by default
                apply_overrides = get_query_param_bool("apply_overrides", default=True)
                resolve_env_vars = get_query_param_bool("resolve_env_vars", default=True)
                # ignored in case of a single config object requested
                include_raw_config = False

            config = await self.deps.config_handler.get_config(config_id, apply_overrides, resolve_env_vars)
            if config:
                headers = {"Resoto-Config-Revision": config.revision}
                if separate_overrides:
                    payload = {"config": config.config, "overrides": self.deps.config_override.get_override(config_id)}
                    if include_raw_config:
                        raw_config = await self.deps.config_handler.get_config(
                            config_id, apply_overrides=False, resolve_env_vars=False
                        )
                        payload["raw_config"] = raw_config.config if raw_config else None
                else:
                    payload = config.config

                return await single_result(request, payload, headers)
            else:
                return not_found

    async def put_config(self, request: Request) -> StreamResponse:
        config_id = ConfigId(request.match_info["config_id"])
        validate = request.query.get("validate", "true").lower() != "false"
        dry_run = request.query.get("dry_run", "false").lower() == "true"
        config = await self.json_from_request(request)
        result = await self.deps.config_handler.put_config(
            ConfigEntity(config_id, config), validate=validate, dry_run=dry_run
        )
        headers = {"Resoto-Config-Revision": result.revision}
        return await single_result(request, result.config, headers)

    async def patch_config(self, request: Request) -> StreamResponse:
        config_id = ConfigId(request.match_info["config_id"])
        validate = request.query.get("validate", "true").lower() != "false"
        dry_run = request.query.get("dry_run", "false").lower() == "true"
        patch = await self.json_from_request(request)
        updated = await self.deps.config_handler.patch_config(
            ConfigEntity(config_id, patch), validate=validate, dry_run=dry_run
        )
        headers = {"Resoto-Config-Revision": updated.revision}
        return await single_result(request, updated.config, headers)

    async def delete_config(self, request: Request) -> StreamResponse:
        config_id = ConfigId(request.match_info["config_id"])
        await self.deps.config_handler.delete_config(config_id)
        return HTTPNoContent()

    async def list_config_models(self, request: Request) -> StreamResponse:
        return await self.stream_response_from_gen(request, self.deps.config_handler.list_config_validation_ids())

    async def get_config_validation(self, request: Request) -> StreamResponse:
        config_id = request.match_info["config_id"]
        model = await self.deps.config_handler.get_config_validation(config_id)
        return await single_result(request, to_js(model)) if model else HTTPNotFound(text="No model for this config.")

    async def get_configs_model(self, request: Request) -> StreamResponse:
        model = await self.deps.config_handler.get_configs_model()
        if request.query.get("flat", "false") == "true":
            model = Model.from_kinds(model.flat_kinds())
        return await single_result(request, to_js(model, strip_nulls=True))

    async def update_configs_model(self, request: Request) -> StreamResponse:
        js = await self.json_from_request(request)
        kinds: List[Kind] = from_js(js, List[Kind])
        model = await self.deps.config_handler.update_configs_model(kinds)
        return await single_result(request, to_js(model))

    async def put_config_validation(self, request: Request) -> StreamResponse:
        config_id = request.match_info["config_id"]
        js = await self.json_from_request(request)
        js["id"] = config_id
        config_model = from_js(js, ConfigValidation)
        model = await self.deps.config_handler.put_config_validation(config_model)
        return await single_result(request, to_js(model))

    async def certificate(self, _: Request) -> StreamResponse:
        cert, fingerprint = self.deps.cert_handler.authority_certificate
        headers = {
            "SHA256-Fingerprint": fingerprint,
            "Content-Disposition": 'attachment; filename="resoto_root_ca.pem"',
        }
        if self.deps.config.args.psk:
            headers["Authorization"] = "Bearer " + encode_jwt(
                {"sha256_fingerprint": fingerprint}, self.deps.config.args.psk
            )
        return HTTPOk(headers=headers, body=cert, content_type="application/x-pem-file")

    async def sign_certificate(self, request: Request) -> StreamResponse:
        csr_bytes = await request.content.read()
        cert, fingerprint = self.deps.cert_handler.sign(csr_bytes)
        headers = {"SHA256-Fingerprint": fingerprint}
        return HTTPOk(headers=headers, body=cert_to_bytes(cert), content_type="application/x-pem-file")

    @staticmethod
    async def metrics(_: Request) -> StreamResponse:
        resp = web.Response(body=prometheus_client.generate_latest())
        resp.content_type = prometheus_client.CONTENT_TYPE_LATEST
        return resp

    async def list_all_subscriptions(self, request: Request) -> StreamResponse:
        subscribers = await self.deps.subscription_handler.all_subscribers()
        return await single_result(request, to_json(subscribers))

    async def get_subscriber(self, request: Request) -> StreamResponse:
        subscriber_id = SubscriberId(request.match_info["subscriber_id"])
        subscriber = await self.deps.subscription_handler.get_subscriber(subscriber_id)
        return self.optional_json(subscriber, f"No subscriber with id {subscriber_id}")

    async def list_subscription_for_event(self, request: Request) -> StreamResponse:
        event_type = request.match_info["event_type"]
        subscribers = await self.deps.subscription_handler.list_subscriber_for(event_type)
        return await single_result(request, to_json(subscribers))

    async def update_subscriber(self, request: Request) -> StreamResponse:
        subscriber_id = SubscriberId(request.match_info["subscriber_id"])
        body = await self.json_from_request(request)
        subscriptions = from_js(body, List[Subscription])
        sub = await self.deps.subscription_handler.update_subscriptions(subscriber_id, subscriptions)
        return await single_result(request, to_json(sub))

    async def delete_subscriber(self, request: Request) -> StreamResponse:
        subscriber_id = SubscriberId(request.match_info["subscriber_id"])
        await self.deps.subscription_handler.remove_subscriber(subscriber_id)
        return web.HTTPNoContent()

    async def add_subscription(self, request: Request) -> StreamResponse:
        subscriber_id = SubscriberId(request.match_info["subscriber_id"])
        event_type = request.match_info["event_type"]
        timeout = timedelta(seconds=int(request.query.get("timeout", "60")))
        wait_for_completion = request.query.get("wait_for_completion", "true").lower() != "false"
        sub = await self.deps.subscription_handler.add_subscription(
            subscriber_id, event_type, wait_for_completion, timeout
        )
        return await single_result(request, to_js(sub))

    async def delete_subscription(self, request: Request) -> StreamResponse:
        subscriber_id = SubscriberId(request.match_info["subscriber_id"])
        event_type = request.match_info["event_type"]
        sub = await self.deps.subscription_handler.remove_subscription(subscriber_id, event_type)
        return await single_result(request, to_js(sub))

    async def handle_subscribed(self, request: Request) -> StreamResponse:
        subscriber_id = SubscriberId(request.match_info["subscriber_id"])
        subscriber = await self.deps.subscription_handler.get_subscriber(subscriber_id)
        if subscriber_id in self.deps.message_bus.active_listener:
            log.info(f"There is already a listener for subscriber: {subscriber_id}. Reject.")
            return web.HTTPTooManyRequests(text="Only one connection per subscriber is allowed!")
        elif subscriber and subscriber.subscriptions:
            pending = await self.deps.task_handler.list_all_pending_actions_for(subscriber)
            return await self.listen_to_events(request, subscriber_id, list(subscriber.subscriptions.keys()), pending)
        else:
            return web.HTTPNotFound(text=f"No subscriber with this id: {subscriber_id} or no subscriptions")

    async def perform_benchmark_on_checks(self, request: Request) -> StreamResponse:
        graph = GraphName(request.match_info["graph_id"])
        provider = request.query.get("provider")
        service = request.query.get("service")
        category = request.query.get("category")
        kind = request.query.get("kind")
        acc = request.query.get("accounts")
        accounts = [a.strip() for a in acc.split(",")] if acc else None
        result = await self.deps.inspector.perform_checks(
            graph, provider=provider, service=service, category=category, kind=kind, accounts=accounts
        )
        return await single_result(request, to_js(result))

    async def perform_benchmark(self, request: Request) -> StreamResponse:
        benchmark = request.match_info["benchmark"]
        graph = GraphName(request.match_info["graph_id"])
        acc = request.query.get("accounts")
        accounts = [a.strip() for a in acc.split(",")] if acc else None
        result = await self.deps.inspector.perform_benchmark(graph, benchmark, accounts=accounts)
        result_graph = result.to_graph()
        async with stream.iterate(result_graph).stream() as streamer:
            await self.stream_response_from_gen(request, streamer, len(result_graph))
        return await single_result(request, to_js(result))

    async def inspection_checks(self, request: Request) -> StreamResponse:
        provider = request.query.get("provider")
        service = request.query.get("service")
        category = request.query.get("category")
        kind = request.query.get("kind")
        inspections = await self.deps.inspector.list_checks(
            provider=provider, service=service, category=category, kind=kind
        )
        return await single_result(request, to_js(inspections))

    async def inspection_results(self, request: Request) -> StreamResponse:
        graph = GraphName(request.match_info["graph_id"])
        check_id = request.match_info["check_id"]
        acc = request.query.get("accounts")
        accounts = [a.strip() for a in acc.split(",")] if acc else None
        inspections = await self.deps.inspector.list_failing_resources(graph, check_id, accounts)
        return await self.stream_response_from_gen(request, inspections)

    async def handle_events(self, request: Request) -> StreamResponse:
        show = request.query["show"].split(",") if "show" in request.query else ["*"]
        return await self.listen_to_events(request, SubscriberId(str(uuid.uuid1())), show)

    async def send_analytics_events(self, request: Request) -> StreamResponse:
        events_json = await self.json_from_request(request)
        events = from_js(events_json, List[AnalyticsEvent])
        await self.deps.event_sender.capture(events)
        return web.HTTPNoContent()

    async def listen_to_events(
        self,
        request: Request,
        listener_id: SubscriberId,
        event_types: List[str],
        initial_messages: Optional[Sequence[Message]] = None,
    ) -> WebSocketResponse:
        handler: Callable[[str], Awaitable[None]] = async_noop

        async def authorize_request(msg: str) -> None:
            nonlocal handler
            if (r := raw_jwt_from_auth_message(msg)) and await self.auth_handler.validate_jwt(r, request) is not None:
                handler = handle_message
            else:
                raise ValueError("No Authorization header provided and no valid auth message sent")

        async def handle_message(msg: str) -> None:
            js = json.loads(msg)
            if "data" in js:
                js["data"]["subscriber_id"] = listener_id
                js["data"]["received_at"] = utc_str()
            message: Message = from_js(js, Message)
            if isinstance(message, Action):
                raise AttributeError("Actors should not emit action messages. ")
            elif isinstance(message, ActionInfo):
                await self.deps.task_handler.handle_action_info(message)
            elif isinstance(message, ActionProgress):
                await self.deps.task_handler.handle_action_progress(message)
            elif isinstance(message, ActionDone):
                await self.deps.task_handler.handle_action_done(message)
            elif isinstance(message, ActionError):
                await self.deps.task_handler.handle_action_error(message)
            else:
                await self.deps.message_bus.emit(message)

        handler = authorize_request if request.get("authorized", False) is False else handle_message
        return await accept_websocket(
            request,
            handle_incoming=lambda x: handler(x),  # pylint: disable=unnecessary-lambda # it is required!
            outgoing_context=partial(self.deps.message_bus.subscribe, listener_id, event_types),
            websocket_handler=self.websocket_handler,
            initial_messages=initial_messages,
        )

    async def handle_work_tasks(self, request: Request) -> WebSocketResponse:
        worker_id = WorkerId(uuid_str())
        worker_descriptions: Future[List[WorkerTaskDescription]] = asyncio.get_event_loop().create_future()
        handler: Callable[[str], Awaitable[None]] = async_noop

        async def authorize_request(msg: str) -> None:
            nonlocal handler
            if (r := raw_jwt_from_auth_message(msg)) and await self.auth_handler.validate_jwt(r, request) is not None:
                handler = handle_connect
            else:
                raise ValueError("No Authorization header provided and no valid auth message sent")

        async def handle_connect(msg: str) -> None:
            nonlocal handler
            cmds = from_js(json.loads(msg), List[WorkerCustomCommand])
            description = [WorkerTaskDescription(cmd.name, cmd.filter) for cmd in cmds]
            # set the future and allow attaching the worker to the task queue
            worker_descriptions.set_result(description)
            # register the descriptions as custom command on the CLI
            for cmd in cmds:
                self.deps.cli.register_alias_template(cmd.to_template())
            # the connect process is done, define the final handler
            handler = handle_message

        async def handle_message(msg: str) -> None:
            tr = from_js(json.loads(msg), WorkerTaskResult)
            if tr.result == "error":
                error = tr.error if tr.error else "worker signalled error without detailed error message"
                await self.deps.worker_task_queue.error_task(worker_id, tr.task_id, error)
            elif tr.result == "done":
                await self.deps.worker_task_queue.acknowledge_task(worker_id, tr.task_id, tr.data)
            else:
                log.info(f"Do not understand this message: {msg}")

        def task_json(task: WorkerTask) -> str:
            return to_js_str(task.to_json())

        @asynccontextmanager
        async def connect_to_task_queue() -> AsyncIterator[Queue[WorkerTask]]:
            # we need to wait for the worker to send the list of commands it can handle
            # before we can attach to the worker task queue
            descriptions = await worker_descriptions
            async with self.deps.worker_task_queue.attach(worker_id, descriptions) as queue:
                yield queue

        handler = authorize_request if request.get("authorized", False) is False else handle_connect
        # noinspection PyTypeChecker
        return await accept_websocket(
            request,
            handle_incoming=lambda x: handler(x),  # pylint: disable=unnecessary-lambda # it is required!
            outgoing_context=connect_to_task_queue,
            websocket_handler=self.websocket_handler,
            outgoing_fn=task_json,
        )

    async def list_work(self, _: Request) -> StreamResponse:
        def wt_to_js(ip: WorkerTaskInProgress) -> Json:
            return {
                "task": ip.task.to_json(),
                "worker": ip.worker.worker_id,
                "retry_counter": ip.retry_counter,
                "deadline": to_json(ip.deadline),
            }

        return web.json_response([wt_to_js(ot) for ot in self.deps.worker_task_queue.outstanding_tasks.values()])

    async def model_uml(self, request: Request) -> StreamResponse:
        output = request.query.get("output", "svg")
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        show = request.query["show"].split(",") if "show" in request.query else None
        hide = request.query["hide"].split(",") if "hide" in request.query else None
        with_inheritance = request.query.get("with_inheritance", "true") != "false"
        with_base_classes = request.query.get("with_base_classes", "true") != "false"
        with_subclasses = request.query.get("with_subclasses", "false") != "false"
        dependency = set(request.query["dependency"].split(",")) if "dependency" in request.query else None
        with_predecessors = request.query.get("with_predecessors", "false") != "false"
        with_successors = request.query.get("with_successors", "false") != "false"
        with_properties = request.query.get("with_properties", "true") != "false"
        aggregate_roots = request.query.get("aggregate_roots", "true") != "false"
        link_classes = request.query.get("link_classes", "false") != "false"
        sort_props = request.query.get("sort_props", "true") != "false"
        result = await self.deps.model_handler.uml_image(
            graph_name=graph_id,
            output=output,
            show_packages=show,
            hide_packages=hide,
            with_inheritance=with_inheritance,
            with_base_classes=with_base_classes,
            with_subclasses=with_subclasses,
            dependency_edges=dependency,  # type: ignore
            with_predecessors=with_predecessors,
            with_successors=with_successors,
            with_properties=with_properties,
            link_classes=link_classes,
            only_aggregate_roots=aggregate_roots,
            sort_props=sort_props,
        )
        response = web.StreamResponse()
        mt = {"svg": "image/svg+xml", "png": "image/png", "puml": "text/plain"}
        response.headers["Content-Type"] = mt[output]
        await response.prepare(request)
        await response.write_eof(result)
        return response

    async def get_model(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        md = await self.deps.model_handler.load_model(graph_id)
        # default to internal model format, but allow to request json schema format
        if request.headers.get("accept") == "application/schema+json":
            return json_response(json_schema(md), content_type="application/schema+json")
        kinds: Iterable[Kind]
        if request.query.get("flat", "false") == "true":
            kinds = md.flat_kinds()
        else:
            kinds = md.kinds.values()
        return await single_result(request, to_js(kinds, strip_nulls=True))

    async def update_model(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        js = await self.json_from_request(request)
        kinds: List[Kind] = from_js(js, List[Kind])
        model = await self.deps.model_handler.update_model(graph_id, kinds)
        return await single_result(request, to_js(model, strip_nulls=True))

    async def get_node(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        node_id = NodeId(request.match_info.get("node_id", "root"))
        graph = self.deps.db_access.get_graph_db(graph_id)
        model = await self.deps.model_handler.load_model(graph_id)
        node = await graph.get_node(model, node_id)
        if node is None:
            return web.HTTPNotFound(text=f"No such node with id {node_id} in graph {graph_id}")
        else:
            return await single_result(request, node)

    async def create_node(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        node_id = NodeId(request.match_info.get("node_id", "some_existing"))
        parent_node_id = NodeId(request.match_info.get("parent_node_id", "root"))
        graph = self.deps.db_access.get_graph_db(graph_id)
        item = await self.json_from_request(request)
        md = await self.deps.model_handler.load_model(graph_id)
        node = await graph.create_node(md, node_id, item, parent_node_id)
        return await single_result(request, node)

    async def update_node(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        node_id = NodeId(request.match_info.get("node_id", "some_existing"))
        section = section_of(request)
        graph = self.deps.db_access.get_graph_db(graph_id)
        patch = await self.json_from_request(request)
        md = await self.deps.model_handler.load_model(graph_id)
        node = await graph.update_node(md, node_id, patch, False, section)
        return await single_result(request, node)

    async def delete_node(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        node_id = NodeId(request.match_info.get("node_id", "some_existing"))
        if node_id == "root":
            raise AttributeError("Root node can not be deleted!")
        graph = self.deps.db_access.get_graph_db(graph_id)
        await graph.delete_node(node_id)
        return web.HTTPNoContent()

    async def update_nodes(self, request: Request) -> StreamResponse:
        graph_name = GraphName(request.match_info.get("graph_id", "resoto"))
        allowed = {*Section.content, "id", "revision"}
        updates: Dict[NodeId, Json] = {}
        async for elem in self.to_json_generator(request):
            keys = set(elem.keys())
            assert keys.issubset(allowed), f"Invalid json. Allowed keys are: {allowed}"
            assert "id" in elem, f"No id given for element {elem}"
            assert keys.intersection(Section.content), f"No update provided for element {elem}"
            uid = elem["id"]
            assert uid not in updates, f"Only one update allowed per id! {elem}"
            del elem["id"]
            updates[uid] = elem
        db = self.deps.db_access.get_graph_db(graph_name)
        model = await self.deps.model_handler.load_model(graph_name)
        result_gen = db.update_nodes(model, updates)
        return await self.stream_response_from_gen(request, result_gen)

    async def list_graphs(self, request: Request) -> StreamResponse:
        graphs = await self.deps.db_access.list_graphs()
        return await single_result(request, graphs)

    async def create_graph(self, request: Request) -> StreamResponse:
        graph_id = request.match_info.get("graph_id", "resoto")
        if "_" in graph_id:
            raise AttributeError("Graph name should not have underscores!")
        graph_name = GraphName(graph_id)
        graph = await self.deps.db_access.create_graph(graph_name)
        model = await self.deps.model_handler.load_model(graph_name)
        root = await graph.get_node(model, NodeId("root"))
        return web.json_response(root)

    async def merge_graph(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        wait_for_result = request.query.get("wait_for_result", "true").lower() == "true"
        task_id: Optional[TaskId] = None
        if tid := request.headers.get("Resoto-Worker-Task-Id"):
            task_id = TaskId(tid)
        log.info(
            f"Received merge_graph request for graph {graph_id}, wait_for_result={wait_for_result}, task_id={task_id}"
        )
        db = self.deps.db_access.get_graph_db(graph_id)
        it = self.to_line_generator(request)
        max_wait = self.deps.config.graph_update.merge_max_wait_time()
        info = await self.deps.graph_merger.merge_graph(db, it, max_wait, None, task_id, wait_for_result)
        return web.json_response(to_js(info)) if info else web.HTTPNoContent()

    async def update_merge_graph_batch(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        wait_for_result = request.query.get("wait_for_result", "true").lower() == "true"
        task_id: Optional[TaskId] = None
        if tid := request.headers.get("Resoto-Worker-Task-Id"):
            task_id = TaskId(tid)
        log.info(f"Received put_sub_graph_batch request for graph {graph_id}, wait_for_result={wait_for_result}")
        db = self.deps.db_access.get_graph_db(graph_id)
        rnd = "".join(SystemRandom().choice(string.ascii_letters) for _ in range(12))
        batch_id = request.query.get("batch_id", rnd)
        it = self.to_line_generator(request)
        max_wait = self.deps.config.graph_update.merge_max_wait_time()
        info = await self.deps.graph_merger.merge_graph(db, it, max_wait, batch_id, task_id, wait_for_result)
        headers = {"BatchId": batch_id}
        return web.json_response(to_json(info), headers=headers) if info else web.HTTPNoContent(headers=headers)

    async def list_batches(self, request: Request) -> StreamResponse:
        graph_db = self.deps.db_access.get_graph_db(GraphName(request.match_info.get("graph_id", "resoto")))
        batch_updates = await graph_db.list_in_progress_updates()
        return web.json_response([b for b in batch_updates if b.get("is_batch")])

    async def commit_batch(self, request: Request) -> StreamResponse:
        graph_db = self.deps.db_access.get_graph_db(GraphName(request.match_info.get("graph_id", "resoto")))
        batch_id = request.match_info.get("batch_id", "some_existing")
        await graph_db.commit_batch_update(batch_id)
        return web.HTTPOk(body="Batch committed.")

    async def abort_batch(self, request: Request) -> StreamResponse:
        graph_db = self.deps.db_access.get_graph_db(GraphName(request.match_info.get("graph_id", "resoto")))
        batch_id = request.match_info.get("batch_id", "some_existing")
        await graph_db.abort_update(batch_id)
        return web.HTTPOk(body="Batch aborted.")

    async def graph_query_model_from_request(self, request: Request) -> Tuple[GraphDB, QueryModel]:
        section = section_of(request)
        query_string = await request.text()
        graph_name = GraphName(request.match_info.get("graph_id", "resoto"))
        raw_at = request.query.get("at")
        at = date_parser.parse(raw_at) if raw_at else None

        snapshot_name = None

        if at:
            snapshot_name = await self.deps.graph_manager.snapshot_at(time=at, graph_name=graph_name)
            if not snapshot_name:
                raise ValueError(f"No snapshot found for {graph_name} at {at}")

        graph_name = snapshot_name or graph_name

        graph_db = self.deps.db_access.get_graph_db(graph_name)
        q = await self.deps.template_expander.parse_query(query_string, section, **request.query)
        m = await self.deps.model_handler.load_model(graph_name)
        return graph_db, QueryModel(q, m)

    async def raw(self, request: Request) -> StreamResponse:
        graph_db, query_model = await self.graph_query_model_from_request(request)
        with_edges = request.query.get("edges") is not None
        query, bind_vars = await graph_db.to_query(query_model, with_edges)
        return web.json_response({"query": query, "bind_vars": bind_vars})

    async def explain(self, request: Request) -> StreamResponse:
        graph_db, query_model = await self.graph_query_model_from_request(request)
        result = await graph_db.explain(query_model)
        return web.json_response(to_js(result))

    async def query_structure(self, request: Request) -> StreamResponse:
        _, query_model = await self.graph_query_model_from_request(request)
        return web.json_response(query_model.query.structure())

    async def query_list(self, request: Request) -> StreamResponse:
        graph_db, query_model = await self.graph_query_model_from_request(request)
        count = request.query.get("count", "true").lower() != "false"
        timeout = if_set(request.query.get("search_timeout"), duration)
        async with await graph_db.search_list(query_model, count, timeout) as cursor:
            return await self.stream_response_from_gen(request, cursor, cursor.count())

    async def cytoscape(self, request: Request) -> StreamResponse:
        graph_db, query_model = await self.graph_query_model_from_request(request)
        result = await graph_db.search_graph(query_model)
        node_link_data = cytoscape_data(result)
        return web.json_response(node_link_data)

    async def query_graph_stream(self, request: Request) -> StreamResponse:
        graph_db, query_model = await self.graph_query_model_from_request(request)
        count = request.query.get("count", "true").lower() != "false"
        timeout = if_set(request.query.get("search_timeout"), duration)
        async with await graph_db.search_graph_gen(query_model, count, timeout) as cursor:
            return await self.stream_response_from_gen(request, cursor, cursor.count())

    async def query_aggregation(self, request: Request) -> StreamResponse:
        graph_db, query_model = await self.graph_query_model_from_request(request)
        async with await graph_db.search_aggregation(query_model) as gen:
            return await self.stream_response_from_gen(request, gen)

    async def query_history(self, request: Request) -> StreamResponse:
        graph_db, query_model = await self.graph_query_model_from_request(request)
        before = request.query.get("before")
        after = request.query.get("after")
        change = request.query.get("change")
        async with await graph_db.search_history(
            query=query_model,
            change=HistoryChange[change] if change else None,
            before=parse_utc(before) if before else None,
            after=parse_utc(after) if after else None,
        ) as gen:
            return await self.stream_response_from_gen(request, gen)

    async def serve_debug_ui(self, request: Request) -> FileResponse:
        """
        This is only for testing different versions of the UI during development.
        """
        commit = request.match_info.get("commit", "default")
        commit = commit[0:6] if len(commit) == 40 else commit  # shorten commit hash
        path = request.match_info.get("path", "index.html")
        dir_path = self.deps.config.run.temp_dir / "ui" / commit
        if not dir_path.exists():
            dir_path.mkdir(parents=True)
            async with self.session.get(f"https://cdn.some.engineering/resoto-ui/commits/{commit}.zip") as resp:
                if resp.status != 200:
                    raise NotFoundError(f"Commit not found: {commit}")
                body = await resp.read()
                with zipfile.ZipFile(BytesIO(body)) as zip_ref:
                    zip_ref.extractall(dir_path)
        file = dir_path / path
        if not file.exists():
            raise NotFoundError(f"File not found: {path}")
        return FileResponse(file)

    async def wipe(self, request: Request) -> StreamResponse:
        graph_id = GraphName(request.match_info.get("graph_id", "resoto"))
        if "truncate" in request.query:
            await self.deps.db_access.get_graph_db(graph_id).wipe()
            return web.HTTPOk(body="Graph truncated.")
        else:
            await self.deps.db_access.delete_graph(graph_id)
            return web.HTTPOk(body="Graph deleted.")

    async def cli_info(self, _: Request) -> StreamResponse:
        def cmd_json(cmd: CLICommand) -> Json:
            return {
                "name": cmd.name,
                "info": cmd.info(),
                "help": cmd.help(),
                "args": to_js(cmd.args_info(), force_dict=True),
                "source": cmd.allowed_in_source_position,
            }

        def alias_json(cmd: AliasTemplate) -> Json:
            return {
                "name": cmd.name,
                "info": cmd.info,
                "help": cmd.help(),
                "args": to_js(cmd.args_info(), force_dict=True),
                "source": cmd.allowed_in_source_position,
            }

        def infra_app_alias_json(cmd: InfraAppAlias) -> Json:
            return {
                "name": cmd.name,
                "info": cmd.description,
                "help": cmd.readme,
                "args": to_js(cmd.parameters, force_dict=True),
                "source": True,
            }

        commands = [
            cmd_json(cmd) for cmd in self.deps.cli.direct_commands.values() if not isinstance(cmd, InternalPart)
        ]
        replacements = self.deps.cli.replacements()
        return web.json_response(
            {
                "commands": commands,
                "replacements": replacements,
                "alias_names": alias_names(),
                "alias_templates": [alias_json(alias) for alias in self.deps.cli.alias_templates.values()],
                "infra_app_aliases": [
                    infra_app_alias_json(alias) for alias in self.deps.cli.infra_app_aliases.values()
                ],
            }
        )

    @staticmethod
    def cli_context_from_request(request: Request) -> CLIContext:
        try:
            columns = int(request.headers.get("Resoto-Shell-Columns", "120"))
            rows = int(request.headers.get("Resoto-Shell-Rows", "50"))
            terminal = request.headers.get("Resoto-Shell-Terminal", "false") == "true"
            colors = ConsoleColorSystem.from_name(request.headers.get("Resoto-Shell-Color-System", "monochrome"))
            renderer = ConsoleRenderer(width=columns, height=rows, color_system=colors, terminal=terminal)
            return CLIContext(env=dict(request.query), console_renderer=renderer, source="api")
        except Exception as ex:
            log.debug("Could not create CLI context.", exc_info=ex)
            return CLIContext(
                env=dict(request.query), console_renderer=ConsoleRenderer.default_renderer(), source="api"
            )

    async def evaluate(self, request: Request) -> StreamResponse:
        ctx = self.cli_context_from_request(request)
        command = await request.text()
        parsed = await self.deps.cli.evaluate_cli_command(command, ctx)

        def line_to_js(line: ParsedCommandLine) -> Json:
            parsed_commands = to_json(line.parsed_commands.commands)
            execute_commands = [{"cmd": part.command.name, "arg": part.arg} for part in line.executable_commands]
            return {"parsed": parsed_commands, "execute": execute_commands, "env": line.parsed_commands.env}

        return web.json_response([line_to_js(line) for line in parsed])

    @timed("api", "execute")
    async def execute(self, request: Request) -> StreamResponse:
        temp_dir: Optional[str] = None
        try:
            ctx = self.cli_context_from_request(request)
            if request.content_type.startswith("text"):
                command = (await request.text()).strip()
            elif request.content_type.startswith("multipart"):
                command = request.headers["Resoto-Shell-Command"].strip()
                temp = tempfile.mkdtemp()
                temp_dir = temp
                files = {}
                # for now, we assume that all multi-parts are file uploads
                async for part in MultipartReader(request.headers, request.content):
                    name = part.name
                    if not name:
                        raise AttributeError("Multipart request: content disposition name is required!")
                    path = os.path.join(temp, rnd_str())  # use random local path to avoid clashes
                    files[name] = path
                    with open(path, "wb") as writer:
                        while not part.at_eof():
                            writer.write(await part.read_chunk())
                ctx = evolve(ctx, uploaded_files=files)
            else:
                raise AttributeError(f"Not able to handle: {request.content_type}")

            # we want to eagerly evaluate the command, so that parse exceptions will throw directly here
            parsed = await self.deps.cli.evaluate_cli_command(command, ctx)
            return await self.execute_parsed(request, command, parsed, ctx)
        finally:
            if temp_dir:
                shutil.rmtree(temp_dir)

    async def execute_parsed(
        self, request: Request, command: str, parsed: List[ParsedCommandLine], ctx: CLIContext
    ) -> StreamResponse:
        # make sure, all requirements are fulfilled
        not_met_requirements = [not_met for line in parsed for not_met in line.unmet_requirements]
        # what is the accepted content type
        # only required for multipart requests
        boundary = "cli-part"
        mp_response = web.StreamResponse(
            status=200, reason="OK", headers={"Content-Type": f"multipart/mixed;boundary={boundary}"}
        )

        if not_met_requirements:
            requirements = [req for line in parsed for cmd in line.executable_commands for req in cmd.action.required]
            data = {"command": command, "env": dict(request.query), "required": to_json(requirements)}
            return web.json_response(data, status=424)
        elif len(parsed) == 1:
            first_result = parsed[0]
            count, generator = await first_result.execute()
            # flat the results from 0 or 1
            async with generator.stream() as streamer:
                gen = await force_gen(streamer)
                if first_result.produces.text:
                    text_gen = ctx.text_generator(first_result, gen)
                    return await self.stream_response_from_gen(request, text_gen, count, first_result.envelope)
                elif first_result.produces.file_path:
                    await mp_response.prepare(request)
                    await Api.multi_file_response(first_result, gen, boundary, mp_response)
                    await Api.close_multi_part_response(mp_response, boundary)
                    return mp_response
                else:
                    raise AttributeError(f"Can not handle type: {first_result.produces}")
        elif len(parsed) > 1:
            await mp_response.prepare(request)
            for single in parsed:
                count, generator = await single.execute()
                async with generator.stream() as streamer:
                    gen = await force_gen(streamer)
                    if single.produces.text:
                        with MultipartWriter(repr(single.produces), boundary) as mp:
                            text_gen = ctx.text_generator(single, gen)
                            content_type, result_stream = await result_binary_gen(request, text_gen)
                            mp.append_payload(
                                AsyncIterablePayload(result_stream, content_type=content_type, headers=single.envelope)
                            )
                            await mp.write(mp_response, close_boundary=False)
                    elif single.produces.file_path:
                        await Api.multi_file_response(single, gen, boundary, mp_response)
                    else:
                        raise AttributeError(f"Can not handle type: {single.produces}")
            await Api.close_multi_part_response(mp_response, boundary)
            return mp_response
        else:
            raise AttributeError("No command could be parsed!")

    @classmethod
    async def json_from_request(cls, request: Request) -> Json:
        if request.content_type in ["application/json"]:
            return await request.json()  # type: ignore
        elif request.content_type in ["application/yaml", "text/yaml"]:
            text = await request.text()
            return yaml.safe_load(text)  # type: ignore
        else:
            raise HTTPNotAcceptable(text="Only support json")

    @classmethod
    async def to_json_generator(cls, request: Request) -> AsyncGenerator[Json, None]:
        async for line in cls.to_line_generator(request):
            yield json.loads(line) if isinstance(line, bytes) else line

    @staticmethod
    def to_line_generator(request: Request) -> AsyncGenerator[Union[bytes, Json], None]:
        async def stream_lines() -> AsyncGenerator[Union[bytes, Json], None]:
            async for line in request.content:
                if len(line.strip()) == 0:
                    continue
                yield line

        async def stream_json_array() -> AsyncGenerator[Union[bytes, Json], None]:
            js_elem = await request.json()
            if isinstance(js_elem, list):
                for doc in js_elem:
                    yield doc
            elif isinstance(js_elem, dict):
                yield js_elem
            else:
                log.warning(f"Received json is neither array nor document: {js_elem}! Ignore.")

        if request.content_type == "application/json":
            return stream_json_array()
        elif request.content_type in ["application/x-ndjson", "application/ndjson"]:
            return stream_lines()
        else:
            raise AttributeError("Can not read graph. Currently supported formats: json and ndjson!")

    @staticmethod
    def optional_json(o: Any, hint: str) -> StreamResponse:
        if o:
            return web.json_response(to_json(o))
        else:
            return web.HTTPNotFound(text=hint)

    @staticmethod
    async def stream_response_from_gen(
        request: Request,
        gen_in: AsyncIterator[JsonElement],
        count: Optional[int] = None,
        additional_header: Optional[Dict[str, str]] = None,
    ) -> StreamResponse:
        # force the async generator, to get an early exception in case of failure
        gen = await force_gen(gen_in)
        content_type, result_gen = await result_binary_gen(request, gen)
        count_header = {"Resoto-Shell-Element-Count": str(count)} if count else {}
        hdr = additional_header or {}
        response = web.StreamResponse(status=200, headers={**hdr, "Content-Type": content_type, **count_header})
        enable_compression(request, response)
        writer: AbstractStreamWriter = await response.prepare(request)  # type: ignore
        cr = "\n".encode("utf-8")
        async for data in result_gen:
            await writer.write(data + cr)
        await response.write_eof()
        return response

    @staticmethod
    async def multi_file_response(
        cmd_line: ParsedCommandLine, results: AsyncIterator[JsonElement], boundary: str, response: StreamResponse
    ) -> None:
        async for file_path in results:
            path = FilePath.from_path(file_path)
            if not (path.local.is_file()):
                raise HTTPNotFound(text=f"No file with this path: {file_path}")
            with open(path.local, "rb") as content:
                headers = cmd_line.envelope
                # only add path header if the user path is more than a file name
                if path.user.name != str(path.user):
                    headers = {**headers, "file-path": str(path.user)}
                with MultipartWriter(boundary=boundary) as mp:
                    pl = BufferedReaderPayload(
                        content,
                        content_type="application/octet-stream",
                        filename=path.user.name,
                        headers=headers,
                    )
                    mp.append_payload(pl)
                    await mp.write(response, close_boundary=False)

    @staticmethod
    async def close_multi_part_response(response: StreamResponse, boundary: str) -> None:
        with MultipartWriter(boundary=boundary) as mp:
            await mp.write(response, close_boundary=True)
        await response.write_eof()



import asyncio
from typing import List

import pytest
from arango.database import StandardDatabase

from resotocore.analytics import AnalyticsEventSender, InMemoryEventSender
from resotocore.db import templatedb
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EventEntityDb
from resotocore.db.templatedb import TemplateEntityDb, EventTemplateEntityDb
from resotocore.query.model import Template


@pytest.fixture
async def template_db(test_db: StandardDatabase) -> TemplateEntityDb:
    async_db = AsyncArangoDB(test_db)
    template_db = templatedb.template_entity_db(async_db, "templates")
    await template_db.create_update_schema()
    await template_db.wipe()
    return template_db


@pytest.fixture
def event_db(template_db: TemplateEntityDb, event_sender: AnalyticsEventSender) -> EventTemplateEntityDb:
    return EventEntityDb(template_db, event_sender, "templates")


@pytest.fixture
def templates() -> List[Template]:
    return [Template(f"tpl_{a}", "is({{a}})") for a in range(0, 10)]


@pytest.mark.asyncio
async def test_load(template_db: TemplateEntityDb, templates: List[Template]) -> None:
    await template_db.update_many(templates)
    loaded = [sub async for sub in template_db.all()]
    assert templates.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_update(template_db: TemplateEntityDb, templates: List[Template]) -> None:
    # multiple updates should work as expected
    await template_db.update_many(templates)
    await template_db.update_many(templates)
    await template_db.update_many(templates)
    loaded = [sub async for sub in template_db.all()]
    assert templates.sort() == loaded.sort()


@pytest.mark.asyncio
async def test_delete(template_db: TemplateEntityDb, templates: List[Template]) -> None:
    await template_db.update_many(templates)
    remaining = list(templates)
    for _ in templates:
        sub = remaining.pop()
        await template_db.delete_value(sub)
        loaded = [sub async for sub in template_db.all()]
        assert remaining.sort() == loaded.sort()
    assert len([sub async for sub in template_db.all()]) == 0


@pytest.mark.asyncio
async def test_events(
    event_db: EventTemplateEntityDb, templates: List[Template], event_sender: InMemoryEventSender
) -> None:
    # 2 times update
    await event_db.update_many(templates)
    await event_db.update_many(templates)
    # 6 times delete
    for sub in templates:
        await event_db.delete_value(sub)
    # make sure all events will arrive
    await asyncio.sleep(0.1)
    # ensure the correct count and order of events
    assert [a.kind for a in event_sender.events] == ["templates-updated-many"] * 2 + ["templates-deleted"] * 10



from resotocore.model.json_schema import json_schema
from resotocore.model.model import Model


def test_schema(foo_model: Model) -> None:
    schema = json_schema(foo_model)
    # The resource is one of the possible 8 types
    assert len(schema["oneOf"]) == 8

    # base type - all properties are defined, additional properties are allowed
    assert schema["$defs"]["base"]["required"] == ["identifier", "kind"]
    assert schema["$defs"]["base"]["properties"].keys() == {"identifier", "kind", "ctime"}
    assert schema["$defs"]["base"]["additionalProperties"] is True

    # final resource type - all properties are defined, additional properties are not allowed
    assert schema["$defs"]["some_complex"]["required"] == []
    expected = {"identifier", "kind", "ctime", "cloud", "account", "parents", "children", "nested"}
    assert schema["$defs"]["some_complex"]["properties"].keys() == expected
    assert schema["$defs"]["some_complex"]["additionalProperties"] is False

    # simple types are defined
    assert schema["$defs"]["datetime"]["type"] == "string"
    assert schema["$defs"]["datetime"]["format"] == "date-time"



import abc
from datetime import datetime
from typing import Union, List

from attrs import define
from deepdiff import DeepDiff
from frozendict import frozendict

from resotocore.model.typed_model import from_js, to_js, to_json
from resotocore.query.model import Query, P
from resotocore.query.query_parser import parse_query
from resotocore.task.task_description import ExecuteOnCLI
from resotocore.types import JsonElement


class ModelBase(abc.ABC):
    def __init__(self, identity: int):
        self.identity = identity


class ModelFoo(ModelBase):
    def __init__(self, identity: int, a: str, b: int):
        super().__init__(identity)
        self.a = a
        self.b = b


class ModelBar(ModelFoo):
    def __init__(self, identity: int, a: str, b: int, foo: str, bla: datetime):
        super().__init__(identity, a, b)
        self.foo = foo
        self.bla = bla


def test_json_marshalling_works() -> None:
    m = ModelFoo(1, "some foo", 23)
    js = to_js(m)
    js["identity"] = 1
    js["a"] = "some foo"
    js["b"] = 23
    again = from_js(js, ModelFoo)
    d = DeepDiff(m, again, truncate_datetime="second")
    assert len(d) == 0


def test_ignore_private_properties() -> None:
    m = ModelFoo(1, "some foo", 23)
    m.__some_private_prop = 23  # type: ignore
    m.__some_other_dunder = "foo"  # type: ignore
    js = to_js(m)
    assert len(js) == 3


def test_marshal_query() -> None:
    q = Query.by("ec2", P("foo") > 23, P("test") >= "bummer", P("das") < "set")
    again = parse_query(str(q))
    assert str(q) == str(again)


def test_frozen_dict() -> None:
    res = ExecuteOnCLI("test", frozendict({"test": "foo"}))
    assert res == from_js(to_js(res), ExecuteOnCLI)


@define
class ClassWithJsonElement:
    elem: JsonElement


@define
class ClassWithStr:
    some_str: str


@define
class ClassWithInt:
    some_int: int


@define
class ClassWithUnion:
    elem: Union[ClassWithInt, ClassWithStr]


def test_json_element() -> None:
    elements = [None, True, False, 1, 1.23, "test", [1, 2, 3], {"test": "foo"}]
    for e in elements:
        t = ClassWithJsonElement(e)
        again = from_js(to_json(t), ClassWithJsonElement)
        assert t == again


def test_union() -> None:
    elements: List[Union[ClassWithInt, ClassWithStr]] = [ClassWithInt(1), ClassWithInt(1), ClassWithStr("test")]
    for e in elements:
        t = ClassWithUnion(e)
        again = from_js(to_json(t), ClassWithUnion)
        assert t == again



import json
import logging
import re
import time
from contextvars import ContextVar
from datetime import datetime, timedelta
from re import RegexFlag
from typing import Any, Dict, Optional, Set, Callable, Awaitable, Tuple, Union
from urllib.parse import urlparse

import jwt
from aiohttp import web
from aiohttp.web import Request, StreamResponse
from aiohttp.web import middleware
from attr import define
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPrivateKey
from cryptography.x509 import Certificate
from jwt import PyJWTError

from resotocore.core_config import CoreConfig
from resotocore.db.system_data_db import SystemDataDb
from resotocore.util import uuid_str
from resotocore.web.certificate_handler import CertificateHandler
from resotolib import jwt as ck_jwt
from resotolib.asynchronous.web import RequestHandler, Middleware
from resotolib.jwt import encode_jwt, create_jwk_dict
from resotolib.types import Json
from resotolib.utils import utc
from resotolib.x509 import gen_rsa_key, gen_csr, key_to_bytes, cert_to_bytes, load_key_from_bytes, load_cert_from_bytes

log = logging.getLogger(__name__)
JWT = Dict[str, Any]
_JWT_Context: ContextVar[JWT] = ContextVar("JWT", default={})
CodeLifeTime = timedelta(minutes=5)


@define
class AuthorizedUser:
    email: str
    roles: Set[str]
    authorized_at: datetime

    def is_valid(self) -> bool:
        return utc() - self.authorized_at < CodeLifeTime


async def jwt_from_context() -> JWT:
    """
    Inside a request handler, this value retrieves the current jwt.
    """
    return _JWT_Context.get()


def raw_jwt_from_auth_message(msg: str) -> Optional[str]:
    """
    Expected message: json object with type kind="authorization" and a jwt field
    { "kind": "authorization", "jwt": "Bearer <jwt>" }
    """
    try:
        js = json.loads(msg)
        assert js.get("kind") == "authorization"
        return js.get("jwt")  # type: ignore
    except Exception:
        return None


@middleware
async def no_check(request: Request, handler: RequestHandler) -> StreamResponse:
    # all requests are authorized automatically
    request["authorized"] = True
    return await handler(request)


class AuthHandler:
    def __init__(
        self,
        system_db: SystemDataDb,
        config: CoreConfig,
        cert_handler: CertificateHandler,
        always_allowed_paths: Set[str],
        not_allowed: Optional[Callable[[Request], Awaitable[StreamResponse]]] = None,
    ) -> None:
        self.system_db = system_db
        self.config = config
        self.psk = config.args.psk
        self.cert_handler = cert_handler
        self.always_allowed_paths = always_allowed_paths
        self.not_allowed = not_allowed
        self.authorization_codes: Dict[str, AuthorizedUser] = {}
        self.signing_key_private: Optional[RSAPrivateKey] = None  # set on start
        self.signing_key_certificate: Optional[Certificate] = None  # set on start
        self.signing_key_jwk: Optional[Json] = None  # set on start

    async def start(self) -> None:
        keys = await self.system_db.jwt_signing_keys()
        # check if the signing key is already in the database
        if keys is None:
            signing_key = gen_rsa_key()
            # TODO: implement certificate renewal
            cert, _ = self.cert_handler.sign(
                gen_csr(signing_key), days_valid=3650, key_usage={"key_cert_sign": True, "key_agreement": True}
            )
            key_string = key_to_bytes(signing_key).decode("utf-8")
            await self.system_db.update_jwt_signing_keys(key_string, cert_to_bytes(cert).decode("utf-8"))
            self.signing_key_private = signing_key
            self.signing_key_certificate = cert
            self.signing_key_jwk = {"keys": [create_jwk_dict(cert)]}
        else:
            key_string, cert_str = keys
            cert = load_cert_from_bytes(cert_str.encode("utf-8"))
            self.signing_key_private = load_key_from_bytes(key_string.encode("utf-8"))
            self.signing_key_certificate = cert
            self.signing_key_jwk = {"keys": [create_jwk_dict(cert)]}

    async def stop(self) -> None:
        pass

    def middleware(self) -> Middleware:
        if self.psk:
            log.info("Use JWT authentication with a pre shared key")
            return self.check_auth()
        else:
            log.info("No authentication requested.")
            return no_check

    async def validate_jwt(self, auth_header: str, request: Request) -> bool:
        def set_valid_jwt(psk_or_cert: Union[str, Certificate, RSAPublicKey]) -> Optional[JWT]:
            try:
                # note: the expiration is already checked by this function
                jwt_token = ck_jwt.decode_jwt_from_header_value(auth_header, psk_or_cert)
            except PyJWTError:
                return None
            if jwt_token:
                request["authorized"] = True  # deferred check in websocket handler
                request["jwt"] = jwt_token
                _JWT_Context.set(jwt_token)
            return jwt_token

        assert self.signing_key_certificate is not None, "AuthHandler not started"
        # based on the jwt, we either use the PSK or the public key
        _, token = auth_header.split(" ", maxsplit=1)
        jwt_header = jwt.get_unverified_header(token)
        # in case of RS256, the public key is used to verify the signature
        secret = self.signing_key_certificate if jwt_header.get("alg") == "RS256" else self.psk
        authorized = set_valid_jwt(secret) is not None
        return authorized

    async def validate_code(self, code: str, request: Request) -> bool:
        if (user := await self.authorized_user(code)) and user.is_valid():
            jwt_token, data = self.user_jwt(user)
            # this will be picked up in on_response_prepare and sent as a header
            request["send_auth_response_header"] = f"Bearer {jwt_token}"
            # set encoded data the same way as if it was a jwt
            request["jwt"] = data
            request["authorized"] = True
            return True
        return False

    def check_auth(self) -> Middleware:
        def always_allowed(request: Request) -> bool:
            for path in self.always_allowed_paths:
                if re.fullmatch(path, request.path, RegexFlag.IGNORECASE):
                    return True
            return False

        @middleware
        async def valid_auth_handler(request: Request, handler: RequestHandler) -> StreamResponse:
            # make sure origin and host match, so the request is valid
            origin: Optional[str] = urlparse(request.headers.get("Origin")).hostname  # type: ignore
            host: Optional[str] = request.headers.get("Host")
            if host is not None and origin is not None:
                if ":" in host:
                    host = host.split(":")[0]
                if origin.lower() != host.lower():
                    log.warning(f"Origin {origin} is not allowed in request from {request.remote} to {request.path}")
                    raise web.HTTPForbidden()

            allowed = False
            # Note: order is important: Authorization header and code is checked even for always allowed paths
            if auth_head := (request.headers.get("Authorization") or request.cookies.get("resoto_authorization")):
                allowed = await self.validate_jwt(auth_head, request)
            elif code := request.query.get("code"):
                allowed = await self.validate_code(code, request)
            elif always_allowed(request):
                allowed = True
            if allowed:
                return await handler(request)
            else:
                if self.not_allowed:
                    return await self.not_allowed(request)
                else:
                    raise web.HTTPUnauthorized()

        return valid_auth_handler

    async def authorized_user(self, code: str) -> Optional[AuthorizedUser]:
        for invalid_code in [k for k, v in self.authorization_codes.items() if not v.is_valid()]:
            self.authorization_codes.pop(invalid_code, None)
        return self.authorization_codes.get(code)

    async def add_authorized_user(self, user: AuthorizedUser) -> str:
        code = str(uuid_str())
        self.authorization_codes[code] = user
        return code

    def user_jwt(self, user: AuthorizedUser) -> Tuple[str, Json]:
        assert self.signing_key_private is not None, "AuthHandler not started"
        assert self.signing_key_certificate is not None, "AuthHandler not started"
        exp = int(time.time() + self.config.api.access_token_expiration_seconds)
        data = {"email": user.email, "roles": ",".join(user.roles), "exp": exp}
        return encode_jwt(data, self.signing_key_private, cert=self.signing_key_certificate), data



from typing import List, Optional, Set, Tuple

from aiohttp import ClientSession
from networkx import MultiDiGraph

from resotocore.cli.model import ParsedCommands, ParsedCommand
from resotocore.config import ConfigValidation
from resotocore.db import EstimatedSearchCost
from resotocore.db.model import GraphUpdate
from resotocore.model.model import Model, Kind
from resotocore.model.typed_model import from_js, to_js
from resotocore.task.model import Subscriber, Subscription
from resotocore.types import Json, JsonElement
from resotocore.ids import NodeId
from resotocore.util import AccessJson


class ApiClient:
    """
    The ApiClient interacts with a running core instance via the REST interface.
    This client is used for testing purposes only.
    """

    def __init__(self, base_path: str, session: ClientSession):
        self.base_path = base_path
        self.session = session

    async def model(self) -> Model:
        async with self.session.get(self.base_path + "/model") as response:
            model_json = await response.json()
            model = Model.from_kinds([from_js(kind, Kind) for kind in model_json])  # type: ignore
            return model

    async def update_model(self, update: List[Kind]) -> Model:
        async with self.session.patch(self.base_path + "/model", json=to_js(update)) as response:
            model_json = await response.json()
            model = Model.from_kinds([from_js(kind, Kind) for kind in model_json["kinds"].values()])  # type: ignore
            return model

    async def list_graphs(self) -> Set[str]:
        async with self.session.get(self.base_path + f"/graph") as response:
            return set(await response.json())

    async def get_graph(self, name: str) -> Optional[AccessJson]:
        async with self.session.post(self.base_path + f"/graph/{name}") as response:
            return AccessJson(await response.json()) if response.status == 200 else None

    async def create_graph(self, name: str) -> AccessJson:
        async with self.session.post(self.base_path + f"/graph/{name}") as response:
            # root node
            return AccessJson(await response.json())

    async def delete_graph(self, name: str, truncate: bool = False) -> str:
        props = {"truncate": "true"} if truncate else {}
        async with self.session.delete(self.base_path + f"/graph/{name}", params=props) as response:
            # root node
            return await response.text()

    async def create_node(self, graph: str, parent_node_id: NodeId, node_id: NodeId, node: Json) -> AccessJson:
        async with self.session.post(
            self.base_path + f"/graph/{graph}/node/{node_id}/under/{parent_node_id}", json=node
        ) as response:
            if response.status == 200:
                return AccessJson(await response.json())
            else:
                raise AttributeError(await response.text())

    async def patch_node(self, graph: str, node_id: NodeId, node: Json, section: Optional[str] = None) -> AccessJson:
        section_path = f"/section/{section}" if section else ""
        async with self.session.patch(
            self.base_path + f"/graph/{graph}/node/{node_id}{section_path}", json=node
        ) as response:
            if response.status == 200:
                return AccessJson(await response.json())
            else:
                raise AttributeError(await response.text())

    async def get_node(self, graph: str, node_id: NodeId) -> AccessJson:
        async with self.session.get(self.base_path + f"/graph/{graph}/node/{node_id}") as response:
            if response.status == 200:
                return AccessJson(await response.json())
            else:
                raise AttributeError(await response.text())

    async def delete_node(self, graph: str, node_id: NodeId) -> None:
        async with self.session.delete(self.base_path + f"/graph/{graph}/node/{node_id}") as response:
            if response.status == 204:
                return None
            else:
                raise AttributeError(await response.text())

    async def patch_nodes(self, graph: str, nodes: List[Json]) -> List[AccessJson]:
        async with self.session.patch(self.base_path + f"/graph/{graph}/nodes", json=nodes) as response:
            if response.status == 200:
                return AccessJson.wrap_list(await response.json())
            else:
                raise AttributeError(await response.text())

    @staticmethod
    def graph_to_json(graph: MultiDiGraph) -> List[Json]:
        ga: List[Json] = [{**node, "type": "node"} for _, node in graph.nodes(data=True)]
        for from_node, to_node, data in graph.edges(data=True):
            ga.append({"type": "edge", "from": from_node, "to": to_node, "edge_type": data["edge_type"]})
        return ga

    async def merge_graph(self, graph: str, update: MultiDiGraph) -> GraphUpdate:
        js = self.graph_to_json(update)
        async with self.session.post(self.base_path + f"/graph/{graph}/merge", json=js) as r:
            if r.status == 200:
                return from_js(await r.json(), GraphUpdate)
            else:
                raise AttributeError(await r.text())

    async def add_to_batch(
        self, graph: str, update: MultiDiGraph, batch_id: Optional[str] = None
    ) -> Tuple[str, GraphUpdate]:
        js = self.graph_to_json(update)
        props = {"batch_id": batch_id} if batch_id else None
        async with self.session.post(self.base_path + f"/graph/{graph}/batch/merge", json=js, params=props) as r:
            if r.status == 200:
                return r.headers["BatchId"], from_js(await r.json(), GraphUpdate)
            else:
                raise AttributeError(await r.text())

    async def list_batches(self, graph: str) -> List[AccessJson]:
        async with self.session.get(self.base_path + f"/graph/{graph}/batch") as r:
            if r.status == 200:
                return AccessJson.wrap_list(await r.json())
            else:
                raise AttributeError(await r.text())

    async def commit_batch(self, graph: str, batch_id: str) -> None:
        async with self.session.post(self.base_path + f"/graph/{graph}/batch/{batch_id}") as r:
            if r.status == 200:
                return None
            else:
                raise AttributeError(await r.text())

    async def abort_batch(self, graph: str, batch_id: str) -> None:
        async with self.session.delete(self.base_path + f"/graph/{graph}/batch/{batch_id}") as r:
            if r.status == 200:
                return None
            else:
                raise AttributeError(await r.text())

    async def search_graph_raw(self, graph: str, search: str) -> AccessJson:
        async with self.session.post(self.base_path + f"/graph/{graph}/search/raw", data=search) as r:
            if r.status == 200:
                return AccessJson.wrap_object(await r.json())
            else:
                raise AttributeError(await r.text())

    async def search_graph_explain(self, graph: str, search: str) -> EstimatedSearchCost:
        async with self.session.post(self.base_path + f"/graph/{graph}/search/explain", data=search) as r:
            if r.status == 200:
                return from_js(await r.json(), EstimatedSearchCost)
            else:
                raise AttributeError(await r.text())

    async def search_list(self, graph: str, search: str) -> List[AccessJson]:
        async with self.session.post(self.base_path + f"/graph/{graph}/search/list", data=search) as r:
            if r.status == 200:
                return AccessJson.wrap_list(await r.json())
            else:
                raise AttributeError(await r.text())

    async def search_graph(self, graph: str, search: str) -> List[AccessJson]:
        async with self.session.post(self.base_path + f"/graph/{graph}/search/graph", data=search) as r:
            if r.status == 200:
                return AccessJson.wrap_list(await r.json())
            else:
                raise AttributeError(await r.text())

    async def search_aggregate(self, graph: str, search: str) -> List[AccessJson]:
        async with self.session.post(self.base_path + f"/graph/{graph}/search/aggregate", data=search) as r:
            if r.status == 200:
                return AccessJson.wrap_list(await r.json())
            else:
                raise AttributeError(await r.text())

    async def subscribers(self) -> List[Subscriber]:
        async with self.session.get(self.base_path + f"/subscribers") as r:
            if r.status == 200:
                return from_js(await r.json(), List[Subscriber])
            else:
                raise AttributeError(await r.text())

    async def subscribers_for_event(self, event_type: str) -> List[Subscriber]:
        async with self.session.get(self.base_path + f"/subscribers/for/{event_type}") as r:
            if r.status == 200:
                return from_js(await r.json(), List[Subscriber])
            else:
                raise AttributeError(await r.text())

    async def subscriber(self, uid: str) -> Optional[Subscriber]:
        async with self.session.get(self.base_path + f"/subscriber/{uid}") as r:
            if r.status == 200:
                return from_js(await r.json(), Subscriber)
            else:
                return None

    async def update_subscriber(self, uid: str, subscriptions: List[Subscription]) -> Optional[Subscriber]:
        async with self.session.put(self.base_path + f"/subscriber/{uid}", json=to_js(subscriptions)) as r:
            if r.status == 200:
                return from_js(await r.json(), Subscriber)
            else:
                raise AttributeError(await r.text())

    async def add_subscription(self, uid: str, subscription: Subscription) -> Subscriber:
        props = {
            "timeout": str(int(subscription.timeout.total_seconds())),
            "wait_for_completion": str(subscription.wait_for_completion),
        }
        async with self.session.post(
            self.base_path + f"/subscriber/{uid}/{subscription.message_type}", params=props
        ) as r:
            if r.status == 200:
                return from_js(await r.json(), Subscriber)
            else:
                raise AttributeError(await r.text())

    async def delete_subscription(self, uid: str, subscription: Subscription) -> Subscriber:
        async with self.session.delete(self.base_path + f"/subscriber/{uid}/{subscription.message_type}") as r:
            if r.status == 200:
                return from_js(await r.json(), Subscriber)
            else:
                raise AttributeError(await r.text())

    async def delete_subscriber(self, uid: str) -> None:
        async with self.session.delete(self.base_path + f"/subscriber/{uid}") as r:
            if r.status == 204:
                return None
            else:
                raise AttributeError(await r.text())

    async def cli_evaluate(self, graph: str, command: str, **env: str) -> List[Tuple[ParsedCommands, List[AccessJson]]]:
        props = {"graph": graph, "section": "reported", **env}
        async with self.session.post(self.base_path + f"/cli/evaluate", data=command, params=props) as r:
            if r.status == 200:
                return [
                    (
                        ParsedCommands(from_js(json["parsed"], List[ParsedCommand]), json["env"]),
                        AccessJson.wrap(json["execute"]),
                    )
                    for json in await r.json()
                ]
            else:
                raise AttributeError(await r.text())

    async def cli_execute(self, graph: str, command: str, **env: str) -> List[JsonElement]:
        props = {"graph": graph, "section": "reported", **env}
        async with self.session.post(self.base_path + f"/cli/execute", data=command, params=props) as r:
            if r.status == 200:
                return AccessJson.wrap_list(await r.json())  # type: ignore
            else:
                raise AttributeError(await r.text())

    async def cli_info(self) -> AccessJson:
        async with self.session.get(self.base_path + f"/cli/info") as r:
            if r.status == 200:
                return AccessJson.wrap_object(await r.json())
            else:
                raise AttributeError(await r.text())

    async def configs(self) -> List[str]:
        async with self.session.get(self.base_path + f"/configs") as r:
            if r.status == 200:
                return AccessJson.wrap_list(await r.json())  # type: ignore
            else:
                raise AttributeError(await r.text())

    async def config(self, config_id: str) -> AccessJson:
        async with self.session.get(self.base_path + f"/config/{config_id}") as r:
            if r.status == 200:
                return AccessJson.wrap_object(await r.json())
            else:
                raise AttributeError(await r.text())

    async def put_config(self, config_id: str, json: Json, validate: bool = True) -> AccessJson:
        params = {"validate": "true" if validate else "false"}
        async with self.session.put(self.base_path + f"/config/{config_id}", json=json, params=params) as r:
            if r.status == 200:
                return AccessJson.wrap_object(await r.json())
            else:
                raise AttributeError(await r.text())

    async def patch_config(self, config_id: str, json: Json) -> AccessJson:
        async with self.session.patch(self.base_path + f"/config/{config_id}", json=json) as r:
            if r.status == 200:
                return AccessJson.wrap_object(await r.json())
            else:
                raise AttributeError(await r.text())

    async def delete_config(self, config_id: str) -> None:
        async with self.session.delete(self.base_path + f"/config/{config_id}") as r:
            if r.status == 204:
                return None
            else:
                raise AttributeError(await r.text())

    async def get_configs_model(self) -> Model:
        async with self.session.get(self.base_path + f"/configs/model") as r:
            if r.status == 200:
                model_json = await r.json()
                model = Model.from_kinds([from_js(kind, Kind) for kind in model_json["kinds"].values()])  # type: ignore
                return model
            else:
                raise AttributeError(await r.text())

    async def update_configs_model(self, update: List[Kind]) -> Model:
        async with self.session.patch(self.base_path + "/configs/model", json=to_js(update)) as response:
            model_json = await response.json()
            model = Model.from_kinds([from_js(kind, Kind) for kind in model_json["kinds"].values()])  # type: ignore
            return model

    async def list_configs_validation(self) -> List[str]:
        async with self.session.get(self.base_path + "/configs/validation") as response:
            return await response.json()  # type: ignore

    async def get_config_validation(self, cfg_id: str) -> Optional[ConfigValidation]:
        async with self.session.get(self.base_path + f"/config_validation/{cfg_id}") as response:
            return from_js(await response.json(), ConfigValidation)

    async def put_config_validation(self, cfg: ConfigValidation) -> ConfigValidation:
        async with self.session.put(self.base_path + f"/config_validation/{cfg.id}", json=to_js(cfg)) as response:
            return from_js(await response.json(), ConfigValidation)

    async def ping(self) -> str:
        async with self.session.get(self.base_path + f"/system/ping") as r:
            if r.status == 200:
                return await r.text()
            else:
                raise AttributeError(await r.text())

    async def ready(self) -> str:
        async with self.session.get(self.base_path + f"/system/ready") as r:
            if r.status == 200:
                return await r.text()
            else:
                raise AttributeError(await r.text())



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, cast

from arango.database import StandardDatabase
from attr import evolve

from resotocore.core_config import CoreConfig
from resotocore.system_start import empty_config, parse_args
from resotocore.types import Json
from resotocore.web.certificate_handler import CertificateHandler
from resotolib.x509 import (
    load_cert_from_bytes,
    cert_fingerprint,
    csr_to_bytes,
    gen_csr,
    gen_rsa_key,
    bootstrap_ca,
    write_cert_to_file,
    write_key_to_file,
)


def test_ca_certificate(cert_handler: CertificateHandler) -> None:
    cert_bytes, fingerprint = cert_handler.authority_certificate
    cert = load_cert_from_bytes(cert_bytes)
    assert cert_fingerprint(cert) == fingerprint


def test_sign(cert_handler: CertificateHandler) -> None:
    cert, fingerprint = cert_handler.sign(csr_to_bytes(gen_csr(gen_rsa_key())))
    assert cert_fingerprint(cert) == fingerprint


def test_bootstrap(test_db: StandardDatabase) -> None:
    with TemporaryDirectory() as tmpdir:
        tmp = Path(tmpdir)
        sd = test_db.collection("system_data")
        config = empty_config()
        # Delete any existing entry, so a new certificate needs to be created
        sd.delete("ca", ignore_missing=True)
        handler = CertificateHandler.lookup(config, test_db, tmp)
        ca = cast(Optional[Json], sd.get("ca"))
        assert ca is not None
        # ensure the certificate in the database is the same as exposed by the handler
        ca_bytes, fingerprint = handler.authority_certificate
        assert ca_bytes == ca["certificate"].encode("utf-8")
        # a new handler will use the existing certificate
        handler2 = CertificateHandler.lookup(config, test_db, tmp)
        assert handler.authority_certificate == handler2.authority_certificate
        # but the host certificate will be different
        assert handler.host_certificate != handler2.host_certificate


def test_load_from_args(default_config: CoreConfig) -> None:
    with TemporaryDirectory() as tmpdir:
        ca_path = tmpdir + "/ca.crt"
        key_path = tmpdir + "/ca.eky"
        pk, cert = bootstrap_ca()
        write_cert_to_file(cert, ca_path)
        write_key_to_file(pk, key_path)
        args = parse_args(["--cert", ca_path, "--cert-key", key_path])
        config = evolve(default_config, args=args)
        context = CertificateHandler._create_host_context(config, cert, pk)
        assert context is not None



import logging
import time
from functools import wraps
from typing import Callable, Any, TypeVar, cast

from prometheus_client import Histogram, Counter, Gauge

# All exported metrics are listed here
Prefix = "resotocore_"
MethodDuration = Histogram(f"{Prefix}method_call_duration", "Duration of single method call", ["module", "name"])
RequestCount = Counter(f"{Prefix}requests_total", "Total Request Count", ["method", "endpoint", "http_status"])
RequestLatency = Histogram(f"{Prefix}request_latency_seconds", "Request latency", ["endpoint"])
RequestInProgress = Gauge(f"{Prefix}requests_in_progress_total", "Requests in progress", ["endpoint", "method"])

# Create a type that is bound to the underlying wrapped function
# This way all signature information is preserved!
DecoratedFn = TypeVar("DecoratedFn", bound=Callable[..., Any])

log = logging.getLogger(__name__)


def perf_now() -> float:
    return time.perf_counter()


def timed(module: str, name: str, is_async: bool = True) -> Callable[[DecoratedFn], DecoratedFn]:
    """
    Use this annotation on a method and measure the duration of the call.
    :param module: the name of the component.
    :param name: the name of the method to be measured.
    :param is_async: set to false if the underlying method is not async
    :return: the wrapped function
    """
    metric = MethodDuration.labels(module=module, name=name)

    def sync_time_wrapper(fn: DecoratedFn) -> DecoratedFn:
        @wraps(fn)
        def async_time_decorated(*args: Any, **kwargs: Any) -> Any:
            start_time = perf_now()
            try:
                rv = fn(*args, **kwargs)
                return rv
            finally:
                duration = perf_now() - start_time
                log.debug(f"Duration of {module}::{name}: {duration}")
                metric.observe(duration)

        return cast(DecoratedFn, async_time_decorated)

    def async_time_wrapper(fn: DecoratedFn) -> DecoratedFn:
        @wraps(fn)
        async def async_time_decorated(*args: Any, **kwargs: Any) -> Any:
            start_time = perf_now()
            try:
                rv = await fn(*args, **kwargs)
                return rv
            finally:
                duration = round((perf_now() - start_time) * 1000)
                log.debug(f"Duration of {module}::{name}: {duration} millis")
                metric.observe(duration)

        return cast(DecoratedFn, async_time_decorated)

    return async_time_wrapper if is_async else sync_time_wrapper



import asyncio
from typing import Optional, AsyncGenerator, Type, Callable, Dict, List, Generic

from resotocore.db.entitydb import EntityDb, T, K
from resotocore.model.typed_model import from_js, to_js
from resotocore.types import Json


# workaround to check generic types in runtime, aka type tags
class TypeTag(Generic[T]):
    def __init__(self, inner: T) -> None:
        self.inner = inner


class InMemoryDb(EntityDb[K, T]):
    def __init__(self, t_type: Type[T], key_fn: Callable[[T], K]):
        self.items: Dict[K, Json] = {}
        self.t_type = t_type
        self.key_fn = key_fn

    async def keys(self) -> AsyncGenerator[K, None]:
        for key in self.items:
            await asyncio.sleep(0)
            yield key

    async def all(self) -> AsyncGenerator[T, None]:
        for js in self.items.values():
            yield from_js(js, self.t_type)

    async def update_many(self, elements: List[T]) -> None:
        for elem in elements:
            key = self.key_fn(elem)
            self.items[key] = to_js(elem)

    async def get(self, key: K) -> Optional[T]:
        js = self.items.get(key)
        return from_js(js, self.t_type) if js else None

    async def update(self, t: T) -> T:
        self.items[self.key_fn(t)] = to_js(t)
        return t

    async def delete(self, key: K) -> bool:
        return self.items.pop(key, None) is not None

    async def delete_value(self, value: T) -> None:
        key = self.key_fn(value)
        self.items.pop(key, None)

    async def create_update_schema(self) -> None:
        pass

    async def wipe(self) -> bool:
        self.items = {}
        return True



from __future__ import annotations

import logging
from asyncio import Lock
from collections import defaultdict
from datetime import timedelta, datetime
from typing import Optional, Iterable, Dict, List

from resotocore.db.subscriberdb import SubscriberDb
from resotocore.message_bus import MessageBus
from resotocore.util import utc, Periodic
from resotocore.task.model import Subscriber, Subscription
from resotocore.ids import SubscriberId
from resotocore.service import Service

log = logging.getLogger(__name__)


class SubscriptionHandler(Service):
    """
    SubscriptionHandler maintains all subscriptions in memory and syncs its internal state with the underlying db.
    Only reason for persistence is recovery of all subscriptions after restart.
    This handler belongs to the event system, which assumes there is only one instance running in each cluster!
    """

    def __init__(self, db: SubscriberDb, message_bus: MessageBus) -> None:
        self.db = db
        self.message_bus = message_bus
        self._subscribers_by_id: Dict[SubscriberId, Subscriber] = {}
        self._subscribers_by_event: Dict[str, List[Subscriber]] = {}
        self.started_at = utc()
        self.cleaner = Periodic("subscription_cleaner", self.check_outdated_handler, timedelta(seconds=10))
        self.not_connected_since: Dict[str, datetime] = {}
        self.lock: Optional[Lock] = None

    async def start(self) -> None:
        self.lock = Lock()
        await self.__load_from_db()
        log.info(f"Loaded {len(self._subscribers_by_id)} subscribers for {len(self._subscribers_by_event)} events")
        await self.cleaner.start()

    async def stop(self) -> None:
        await self.cleaner.stop()

    async def all_subscribers(self) -> Iterable[Subscriber]:
        return self._subscribers_by_id.values()

    async def get_subscriber(self, subscriber_id: SubscriberId) -> Optional[Subscriber]:
        return self._subscribers_by_id.get(subscriber_id)

    async def list_subscriber_for(self, event_type: str) -> List[Subscriber]:
        return self._subscribers_by_event.get(event_type, [])

    async def add_subscription(
        self, subscriber_id: SubscriberId, event_type: str, wait_for_completion: bool, timeout: timedelta
    ) -> Subscriber:
        existing = self._subscribers_by_id.get(subscriber_id, Subscriber(subscriber_id, {}))
        updated = existing.add_subscription(event_type, wait_for_completion, timeout)
        if existing != updated:
            log.info(f"Subscriber {subscriber_id}: add subscription={event_type} ({wait_for_completion}, {timeout})")
            await self.db.update(updated)
            await self.__load_from_db()
        return updated

    async def remove_subscription(self, subscriber_id: SubscriberId, event_type: str) -> Subscriber:
        existing = self._subscribers_by_id.get(subscriber_id, Subscriber(subscriber_id, {}))
        updated = existing.remove_subscription(event_type)
        if existing != updated:
            log.info(f"Subscriber {subscriber_id}: remove subscription={event_type}")
            if updated.subscriptions:
                await self.db.update(updated)
            else:
                await self.db.delete(subscriber_id)
            await self.__load_from_db()
        return updated

    async def update_subscriptions(self, subscriber_id: SubscriberId, subscriptions: List[Subscription]) -> Subscriber:
        existing = self._subscribers_by_id.get(subscriber_id, None)
        updated = Subscriber.from_list(subscriber_id, subscriptions)
        if existing != updated:
            log.info(f"Subscriber {subscriber_id}: update all subscriptions={subscriptions}")
            await self.db.update(updated)
            await self.__load_from_db()
        return updated

    async def remove_subscriber(self, subscriber_id: SubscriberId) -> Optional[Subscriber]:
        existing = self._subscribers_by_id.get(subscriber_id, None)
        if existing:
            log.info(f"Subscriber {subscriber_id}: remove subscriber")
            await self.db.delete(subscriber_id)
            await self.__load_from_db()
        return existing

    async def __load_from_db(self) -> None:
        assert self.lock is not None
        async with self.lock:
            self._subscribers_by_id = {s.id: s async for s in self.db.all()}
            self._subscribers_by_event = self.update_subscriber_by_event(self._subscribers_by_id.values())

    def subscribers_by_event(self) -> Dict[str, List[Subscriber]]:
        return self._subscribers_by_event

    @staticmethod
    def update_subscriber_by_event(subscribers: Iterable[Subscriber]) -> Dict[str, List[Subscriber]]:
        result: Dict[str, List[Subscriber]] = defaultdict(list)
        for subscriber in subscribers:
            for subscription in subscriber.subscriptions.values():
                result[subscription.message_type].append(subscriber)
        return result

    async def check_outdated_handler(self) -> None:
        """
        Periodically check, if there are subscribers that have subscribed, but are not connected.
        The subscription will be removed when a dead subscription is detected.
        """
        now = utc()
        # In case the service has just been started/restarted:
        # do not remove any subscriptions during the first minutes.
        if (now - self.started_at) > timedelta(minutes=5):
            expected = set(self._subscribers_by_id.keys())
            connected = set(self.message_bus.active_listener.keys())
            # remove all connected subscriber from the not connected map
            for c in connected:
                self.not_connected_since.pop(c, None)
            missing = expected - connected
            for subscriber in missing:
                at = self.not_connected_since.get(subscriber)
                if at and (now - at) > timedelta(minutes=3):
                    log.warning(f"Subscriber {subscriber} is missing. Remove all subscription.")
                    await self.remove_subscriber(subscriber)
                    self.not_connected_since.pop(subscriber, None)
                elif at:
                    pass
                else:
                    self.not_connected_since[subscriber] = now



from textwrap import dedent

from resotocore.cli.model import (
    CLIContext,
    AliasTemplate,
    AliasTemplateParameter,
    InfraAppAlias,
    InfraAppAliasParameter,
)
from resotocore.console_renderer import ConsoleRenderer, ConsoleColorSystem


def test_format() -> None:
    context = CLIContext()
    fn = context.formatter("foo={foo} and bla={bla}: {bar}")
    assert fn({}) == "foo=null and bla=null: null"
    assert fn({"foo": 1, "bla": 2, "bar": 3}) == "foo=1 and bla=2: 3"


def test_context_format() -> None:
    context = CLIContext()
    fn, vs = context.formatter_with_variables("foo={foo} and bla={bla}: {bar}")
    assert vs == {"foo", "bla", "bar"}
    assert fn({}) == "foo=null and bla=null: null"
    assert fn({"foo": 1, "bla": 2, "bar": 3}) == "foo=1 and bla=2: 3"


def test_supports_color() -> None:
    assert not CLIContext().supports_color()
    assert not CLIContext(console_renderer=ConsoleRenderer()).supports_color()
    assert not CLIContext(console_renderer=ConsoleRenderer(color_system=ConsoleColorSystem.monochrome)).supports_color()
    assert CLIContext(console_renderer=ConsoleRenderer(color_system=ConsoleColorSystem.standard)).supports_color()
    assert CLIContext(console_renderer=ConsoleRenderer(color_system=ConsoleColorSystem.eight_bit)).supports_color()
    assert CLIContext(console_renderer=ConsoleRenderer(color_system=ConsoleColorSystem.truecolor)).supports_color()


def test_alias_template() -> None:
    params = [AliasTemplateParameter("a", "some a"), AliasTemplateParameter("b", "some b", "bv")]
    tpl = AliasTemplate("foo", "does foes", "{{a}} | {{b}}", params)
    assert tpl.render({"a": "test", "b": "bla"}) == "test | bla"
    assert CLIContext().render_console(tpl.help()) == dedent(
        """
        foo: does foes
        ```shell
        foo --a <value> --b <value>
        ```

        ## Parameters
        - `--a` [required]: some a
        - `--b` [default: bv]: some b

        ## Template
        ```shell
        > {{a}} | {{b}}
        ```

        ## Example
        ```shell
        # Executing this command
        > foo --a "test-a"
        # Will expand to this command
        > test-a | bv
        ```
        """
    )

    tpl_no_args = AliasTemplate(
        "bla",
        "does blas",
        "bla {{args}}",
        description="This is doing bla",
        args_description={"p1": "is doing awesome", "p2": "enables after burners"},
    )
    assert tpl_no_args.render({"args": "something"}) == "bla something"
    assert (
        CLIContext().render_console(tpl_no_args.help()).strip()
        == dedent(
            """
            bla: does blas
            ```shell
            bla  [p1] [p2]
            ```

            ## Parameters

            - `p1`: is doing awesome
            - `p2`: enables after burners

            This is doing bla
            """
        ).strip()
    )


def test_infra_app_alias() -> None:
    params = [
        InfraAppAliasParameter(
            name="param_a",
            help="some a",
            default=None,
        ),
        InfraAppAliasParameter(
            name="param_b",
            help="some b",
            default="default_b",
        ),
    ]
    alias = InfraAppAlias("foo", "does foes", "readme", params)
    assert alias.render({"args": "args_go_here"}) == "apps run foo args_go_here"
    assert alias.rendered_help(CLIContext()) == dedent(
        """
        foo: does foes
        ```shell
        foo --param-a <value> --param-b <value>
        ```

        readme
        ## Parameters
        - `--param-a` [required]: some a
        - `--param-b` [default: default_b]: some b"""
    )



import logging
from argparse import Namespace
from datetime import datetime, timezone, timedelta
from time import sleep
from typing import Dict, List, Tuple, Union, cast, Optional

from arango import ArangoServerError
from arango.client import ArangoClient
from arango.database import StandardDatabase
from dateutil.parser import parse
from requests.exceptions import RequestException

from resotocore.analytics import AnalyticsEventSender
from resotocore.async_extensions import run_async
from resotocore.core_config import CoreConfig
from resotocore.db import SystemData
from resotocore.db.arangodb_extensions import ArangoHTTPClient
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.configdb import config_entity_db, config_validation_entity_db
from resotocore.db.deferred_edge_db import pending_deferred_edge_db
from resotocore.db.entitydb import EventEntityDb
from resotocore.db.graphdb import ArangoGraphDB, GraphDB, EventGraphDB
from resotocore.db.jobdb import job_db
from resotocore.db.modeldb import ModelDb, model_db
from resotocore.db.packagedb import app_package_entity_db
from resotocore.db.runningtaskdb import running_task_db
from resotocore.db.subscriberdb import subscriber_db
from resotocore.db.system_data_db import SystemDataDb
from resotocore.db.templatedb import template_entity_db
from resotocore.error import NoSuchGraph, RequiredDependencyMissingError
from resotocore.ids import GraphName
from resotocore.model.adjust_node import AdjustNode
from resotocore.model.graph_access import EdgeTypes
from resotocore.model.typed_model import from_js, to_js
from resotocore.service import Service
from resotocore.types import Json
from resotocore.util import Periodic, utc, shutdown_process, uuid_str, check_graph_name

log = logging.getLogger(__name__)


class DbAccess(Service):
    def __init__(
        self,
        arango_database: StandardDatabase,
        event_sender: AnalyticsEventSender,
        adjust_node: AdjustNode,
        config: CoreConfig,
        subscriber_name: str = "subscribers",
        running_task_name: str = "running_tasks",
        job_name: str = "jobs",
        deferred_edge_name: str = "deferred_outer_edges",
        config_entity: str = "configs",
        config_validation_entity: str = "config_validation",
        configs_model: str = "configs_model",
        template_entity: str = "templates",
        infra_app_packages: str = "infra_app_packages",
    ):
        self.event_sender = event_sender
        self.database = arango_database
        self.db = AsyncArangoDB(arango_database)
        self.adjust_node = adjust_node
        self.graph_model_dbs: Dict[GraphName, ModelDb] = {}
        self.subscribers_db = EventEntityDb(subscriber_db(self.db, subscriber_name), event_sender, subscriber_name)
        self.system_data_db = SystemDataDb(self.db)
        self.running_task_db = running_task_db(self.db, running_task_name)
        self.pending_deferred_edge_db = pending_deferred_edge_db(self.db, deferred_edge_name)
        self.job_db = job_db(self.db, job_name)
        self.config_entity_db = config_entity_db(self.db, config_entity)
        self.config_validation_entity_db = config_validation_entity_db(self.db, config_validation_entity)
        self.configs_model_db = model_db(self.db, configs_model)
        self.template_entity_db = template_entity_db(self.db, template_entity)
        self.package_entity_db = app_package_entity_db(self.db, infra_app_packages)
        self.graph_dbs: Dict[str, GraphDB] = {}
        self.config = config
        self.cleaner = Periodic("outdated_updates_cleaner", self.check_outdated_updates, timedelta(seconds=60))

    async def start(self) -> None:
        await self.subscribers_db.create_update_schema()
        await self.running_task_db.create_update_schema()
        await self.job_db.create_update_schema()
        await self.config_entity_db.create_update_schema()
        await self.config_validation_entity_db.create_update_schema()
        await self.configs_model_db.create_update_schema()
        await self.template_entity_db.create_update_schema()
        await self.pending_deferred_edge_db.create_update_schema()
        await self.package_entity_db.create_update_schema()
        for graph in cast(List[Json], self.database.graphs()):
            graph_name = GraphName(graph["name"])

            # snapshot graphs do not need any schema migrations,
            # we can skip them
            if str(graph_name).startswith("snapshot"):
                continue

            log.info(f"Found graph: {graph_name}")
            db = self.get_graph_db(graph_name)
            await db.create_update_schema()
            await self.get_graph_model_db(graph_name)
        await self.cleaner.start()

    async def stop(self) -> None:
        await self.cleaner.stop()

    def graph_model_name(self, graph_name: GraphName) -> str:
        return f"{graph_name}_model"

    async def create_graph(self, name: GraphName, validate_name: bool = True) -> GraphDB:
        if validate_name:
            check_graph_name(name)

        db = self.get_graph_db(name, no_check=True)
        await db.create_update_schema()
        return db

    async def delete_graph(self, name: GraphName) -> None:
        def delete() -> None:
            db = self.database
            if db.has_graph(name):
                # delete arrangodb graph
                db.delete_graph(name, drop_collections=True, ignore_missing=True)
                # delete vertex collections just in case
                db.delete_collection(name, ignore_missing=True)
                # delete edge collections
                for edge_type in EdgeTypes.all:
                    db.delete_collection(f"{name}_{edge_type}", ignore_missing=True)
                # delete the rest
                db.delete_collection(f"{name}_in_progress", ignore_missing=True)
                db.delete_view(f"search_{name}", ignore_missing=True)
                # remove all temp collection names
                for coll in cast(List[Json], db.collections()):
                    if coll["name"].startswith(f"{name}_temp_"):
                        db.delete_collection(coll["name"], ignore_missing=True)
                self.graph_dbs.pop(name, None)

        return await run_async(delete)

    async def delete_graph_model(self, graph_name: GraphName) -> None:
        await self.db.delete_collection(self.graph_model_name(graph_name), ignore_missing=True)
        self.graph_model_dbs.pop(graph_name, None)

    async def list_graphs(self) -> List[GraphName]:
        return [a["name"] for a in cast(List[Json], self.database.graphs()) if not a["name"].endswith("_hs")]

    def get_graph_db(self, name: GraphName, no_check: bool = False) -> GraphDB:
        if name in self.graph_dbs:
            return self.graph_dbs[name]
        else:
            if not no_check and not self.database.has_graph(name):
                raise NoSuchGraph(name)
            graph_db = ArangoGraphDB(self.db, name, self.adjust_node, self.config.graph_update)
            event_db = EventGraphDB(graph_db, self.event_sender)
            self.graph_dbs[name] = event_db
            return event_db

    async def get_graph_model_db(self, graph_name: GraphName) -> ModelDb:
        if db := self.graph_model_dbs.get(graph_name):
            return db
        else:
            model_name = self.graph_model_name(graph_name)
            db = EventEntityDb(model_db(self.db, model_name), self.event_sender, model_name)
            await db.create_update_schema()
            self.graph_model_dbs[graph_name] = db
            return db

    async def check_outdated_updates(self) -> None:
        now = datetime.now(timezone.utc)
        for db in self.graph_dbs.values():
            for update in await db.list_in_progress_updates():
                created = datetime.fromtimestamp(parse(update["created"]).timestamp(), timezone.utc)
                if (now - created) > self.config.graph_update.abort_after():
                    batch_id = update["id"]
                    log.warning(f"Given update is too old: {batch_id}. Will abort the update.")
                    await db.abort_update(batch_id)

    # Only used during startup.
    # Note: this call uses sleep and will block the current executing thread!
    @classmethod
    def connect(
        cls, args: Namespace, timeout: timedelta, sleep_time: float = 5, verify: Union[str, bool, None] = None
    ) -> Tuple[bool, SystemData, StandardDatabase]:
        deadline = utc() + timeout
        db = cls.client(args, verify)

        def create_database() -> None:
            try:
                # try to access the system database with default credentials.
                # this only works if arango has been started with default settings.
                http_client = ArangoHTTPClient(args.graphdb_request_timeout, not args.graphdb_no_ssl_verify)
                root_pw = args.graphdb_root_password
                secure_root = not args.graphdb_bootstrap_do_not_secure
                root_db = ArangoClient(hosts=args.graphdb_server, http_client=http_client).db(password=root_pw)
                root_db.echo()  # this call will fail, if we are not allowed to access the system db
                user = args.graphdb_username
                passwd = args.graphdb_password
                database = args.graphdb_database
                change = False
                if not root_db.has_user(user):
                    log.info("Configured graph db user does not exist. Create it.")
                    root_db.create_user(user, passwd, active=True)
                    change = True
                if not root_db.has_database(database):
                    log.info("Configured graph db database does not exist. Create it.")
                    root_db.create_database(
                        database,
                        [{"username": user, "password": passwd, "active": True, "extra": {"generated": "resoto"}}],
                    )
                    change = True
                if change and secure_root and root_pw == "" and passwd != "" and passwd not in {"test"}:
                    root_db.replace_user("root", passwd, True)
                    log.info(
                        "Database is using an empty password. "
                        "Secure the root account with the provided user password. "
                        "Login to the Resoto database via provided username and password. "
                        "Login to the System database via `root` and provided password!"
                    )
                if not change:
                    log.info("Not allowed to access database, while user and database exist. Wrong password?")
            except Exception as ex:
                log.error(
                    "Database or user does not exist or does not have enough permissions. "
                    f"Attempt to create user/database via default system account is not possible. Reason: {ex}. "
                    "You can provide the password of the root user via --graphdb-root-password to setup "
                    "a Resoto user and database automatically."
                )

        def system_data() -> Tuple[bool, SystemData]:
            def insert_system_data() -> SystemData:
                system = SystemData(uuid_str(), utc(), 1)
                log.info(f"Create new system data entry: {system}")
                db.insert_document("system_data", {"_key": "system", **to_js(system)}, overwrite=True)
                return system

            if not db.has_collection("system_data"):
                db.create_collection("system_data")

            sys_js: Optional[Json] = db.collection("system_data").get("system")  # type: ignore
            return (True, insert_system_data()) if not sys_js else (False, from_js(sys_js, SystemData))

        while True:
            try:
                db.echo()
                try:
                    db_version = int(cast(str, db.required_db_version()))
                except Exception as ex:
                    log.warning(f"Not able to retrieve version of arangodb. Reason: {ex}. Continue.")
                else:
                    if db_version < 30802:
                        raise RequiredDependencyMissingError("Need arangodb in version 3.8.2 or later")

                created, sys_data = system_data()
                return created, sys_data, db
            except ArangoServerError as ex:
                if utc() > deadline:
                    log.error("Can not connect to database. Giving up.")
                    shutdown_process(1)
                elif ex.error_code in (11, 1228, 1703):
                    # https://www.arangodb.com/docs/stable/appendix-error-codes.html
                    # This means we can reach the database, but are either not allowed to access it
                    # or the related user and or database could not be found.
                    # We assume the database does not exist and try to create it.
                    create_database()
                else:
                    log.warning(f"Problem accessing the graph database: {ex}. Trying again in 5 seconds.")
                # Retry directly after the first attempt
                sleep(sleep_time)
            except (RequestException, ConnectionError) as ex:
                log.warning(f"Can not access database. Trying again in 5 seconds: {ex}")
                sleep(sleep_time)

    @staticmethod
    def client(args: Namespace, verify: Union[bool, str, None] = None) -> StandardDatabase:
        if args.graphdb_type not in "arangodb":
            log.fatal(f"Unknown Graph DB type {args.graphdb_type}")
            shutdown_process(1)

        http_client = ArangoHTTPClient(args.graphdb_request_timeout, verify=verify)
        client = ArangoClient(hosts=args.graphdb_server, http_client=http_client)
        return client.db(args.graphdb_database, username=args.graphdb_username, password=args.graphdb_password)



import asyncio
import shutil
import tempfile
from asyncio import Queue
from collections import defaultdict
from contextlib import suppress
from datetime import timedelta
from pathlib import Path
from tempfile import TemporaryDirectory
from types import SimpleNamespace
from typing import AsyncGenerator, Iterator, Dict, Any, Generator, Callable
from typing import List, Optional
from typing import Tuple, AsyncIterator, cast

from aiohttp import ClientSession
from aiohttp.hdrs import METH_ANY
from aiohttp.test_utils import TestServer
from aiohttp.web import Request, Response, Application, route
from arango.client import ArangoClient
from arango.database import StandardDatabase
from attr import evolve
from pytest import fixture
from rich.console import Console

from resotocore.action_handlers.merge_outer_edge_handler import MergeOuterEdgesHandler
from resotocore.analytics import AnalyticsEventSender, InMemoryEventSender, NoEventSender
from resotocore.cli.cli import CLIService
from resotocore.cli.command import (
    alias_names,
    all_commands,
)
from resotocore.dependencies import Dependencies
from resotocore.config import ConfigHandler, ConfigEntity, ConfigValidation, ConfigOverride
from resotocore.config.config_handler_service import ConfigHandlerService
from resotocore.config.core_config_handler import CoreConfigHandler
from resotocore.console_renderer import ConsoleRenderer, ConsoleColorSystem
from resotocore.core_config import (
    GraphUpdateConfig,
    CoreConfig,
    EditableConfig,
    DatabaseConfig,
    RuntimeConfig,
    CustomCommandsConfig,
    SnapshotsScheduleConfig,
    RunConfig,
)
from resotocore.db import runningtaskdb, SystemData
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.db_access import DbAccess
from resotocore.db.graphdb import ArangoGraphDB, EventGraphDB
from resotocore.db.jobdb import JobDb
from resotocore.db.packagedb import PackageEntityDb, app_package_entity_db
from resotocore.db.runningtaskdb import RunningTaskDb
from resotocore.db.system_data_db import SystemDataDb
from resotocore.graph_manager.graph_manager import GraphManager
from resotocore.system_start import empty_config, parse_args
from resotocore.ids import SubscriberId, WorkerId, TaskDescriptorId, ConfigId, GraphName
from resotocore.infra_apps.local_runtime import LocalResotocoreAppRuntime
from resotocore.infra_apps.package_manager import PackageManager
from resotocore.message_bus import (
    MessageBus,
    Message,
    Event,
    Action,
    ActionDone,
)
from resotocore.model.adjust_node import NoAdjust
from resotocore.model.db_updater import GraphMerger
from resotocore.model.graph_access import EdgeTypes, Section
from resotocore.model.model import Model, Kind, ComplexKind, Property, SyntheticProperty, StringKind
from resotocore.model.resolve_in_graph import GraphResolver
from resotocore.model.resolve_in_graph import NodePath
from resotocore.model.typed_model import to_js
from resotocore.query.template_expander import TemplateExpander
from resotocore.report import BenchmarkConfigPrefix, CheckConfigPrefix, Benchmark
from resotocore.report.inspector_service import InspectorService
from resotocore.report.report_config import BenchmarkConfig
from resotocore.task.task_dependencies import TaskDependencies
from resotocore.task.model import Subscriber, Subscription
from resotocore.task.scheduler import Scheduler
from resotocore.task.subscribers import SubscriptionHandler
from resotocore.task.task_description import (
    Step,
    PerformAction,
    WaitForEvent,
    EmitEvent,
    StepErrorBehaviour,
    EventTrigger,
    RunningTask,
    ExecuteCommand,
    TaskSurpassBehaviour,
    Job,
    Workflow,
    TimeTrigger,
    WaitForCollectDone,
)
from resotocore.task.task_handler import TaskHandlerService
from resotocore.types import Json
from resotocore.user import UserManagement
from resotocore.user.user_management import UserManagementService
from resotocore.util import value_in_path, uuid_str, utc
from resotocore.web.certificate_handler import CertificateHandler
from resotocore.worker_task_queue import WorkerTaskQueue, WorkerTaskDescription, WorkerTask, WorkerTaskName
from resotolib.x509 import bootstrap_ca
from tests.resotocore import create_graph
from tests.resotocore.db.entitydb import InMemoryDb
from tests.resotocore.model import ModelHandlerStatic
from tests.resotocore.query.template_expander_test import InMemoryTemplateExpander


@fixture
def default_config() -> CoreConfig:
    ed = EditableConfig()
    return CoreConfig(
        api=ed.api,
        cli=ed.cli,
        db=DatabaseConfig(),
        graph_update=ed.graph_update,
        # We use this flag explicitly - otherwise it is picked up by env vars
        runtime=RuntimeConfig(usage_metrics=False),
        workflows=ed.workflows,
        custom_commands=CustomCommandsConfig(),
        snapshots=SnapshotsScheduleConfig(),
        args=parse_args(["--analytics-opt-out"]),
        run=RunConfig(),
    )


@fixture
def message_bus() -> MessageBus:
    return MessageBus()


@fixture
def job_db() -> JobDb:
    return InMemoryDb[str, Job](Job, lambda x: x.id)


@fixture
def event_sender() -> InMemoryEventSender:
    return InMemoryEventSender()


@fixture
def local_client() -> ArangoClient:
    return ArangoClient(hosts="http://localhost:8529")


@fixture
def system_db(local_client: ArangoClient) -> StandardDatabase:
    return local_client.db()


@fixture
async def async_db(test_db: StandardDatabase) -> AsyncArangoDB:
    return AsyncArangoDB(test_db)


@fixture
async def all_events(message_bus: MessageBus) -> AsyncGenerator[List[Message], None]:
    events: List[Message] = []

    async def gather_events() -> None:
        async with message_bus.subscribe(SubscriberId("test")) as event_queue:
            while True:
                events.append(await event_queue.get())

    run_gather = asyncio.create_task(gather_events())
    try:
        yield events
    finally:
        run_gather.cancel()


@fixture
def test_db(local_client: ArangoClient, system_db: StandardDatabase) -> StandardDatabase:
    if not system_db.has_user("test"):
        system_db.create_user("test", "test", True)

    if not system_db.has_database("test"):
        system_db.create_database("test", [{"username": "test", "password": "test", "active": True}])

    # Connect to "test" database as "test" user.
    return local_client.db("test", username="test", password="test")


@fixture
async def graph_db(async_db: AsyncArangoDB) -> ArangoGraphDB:
    graph_db = ArangoGraphDB(async_db, GraphName("ns"), NoAdjust(), GraphUpdateConfig())
    await graph_db.create_update_schema()
    await async_db.truncate(graph_db.in_progress)
    return graph_db


@fixture
async def running_task_db(async_db: AsyncArangoDB) -> RunningTaskDb:
    task_db = runningtaskdb.running_task_db(async_db, "running_tasks")
    await task_db.create_update_schema()
    await task_db.wipe()
    return task_db


@fixture()
def db_access(graph_db: ArangoGraphDB) -> DbAccess:
    access = DbAccess(graph_db.db.db, NoEventSender(), NoAdjust(), empty_config())
    return access


@fixture
async def package_entity_db(async_db: AsyncArangoDB) -> PackageEntityDb:
    package_entity_db = app_package_entity_db(async_db, "test_package_entity_db")
    await package_entity_db.create_update_schema()
    await package_entity_db.wipe()
    return package_entity_db


@fixture
def foo_kinds() -> List[Kind]:
    base = ComplexKind(
        "base",
        [],
        [
            Property("identifier", "string", required=True),
            Property("kind", "string", required=True),
            Property("ctime", "datetime"),
        ],
        aggregate_root=False,
    )
    foo = ComplexKind(
        "foo",
        ["base"],
        [
            Property("name", "string"),
            Property("some_int", "int32"),
            Property("some_string", "string"),
            Property("now_is", "datetime"),
            Property("ctime", "datetime"),
            Property("age", "trafo.duration_to_datetime", False, SyntheticProperty(["ctime"])),
        ],
        successor_kinds={EdgeTypes.default: ["bla"]},
    )
    inner = ComplexKind("inner", [], [Property("name", "string"), Property("inner", "inner[]")], aggregate_root=False)
    bla = ComplexKind(
        "bla",
        ["base"],
        [
            Property("name", "string"),
            Property("now", "date"),
            Property("f", "int32"),
            Property("g", "int32[]"),
            Property("h", "inner"),
        ],
        successor_kinds={EdgeTypes.default: ["bla"]},
    )
    cloud = ComplexKind("cloud", ["foo"], [Property("nick", "string")])
    account = ComplexKind("account", ["foo"], [Property("nick", "string")])
    region = ComplexKind("region", ["foo"], [Property("nick", "string")])
    parent = ComplexKind("parent", ["foo"], [Property("nick", "string")])
    child = ComplexKind("child", ["foo"], [Property("nick", "string")])
    some_complex = ComplexKind(
        "some_complex",
        ["base"],
        [
            Property("cloud", "cloud"),
            Property("account", "account"),
            Property("parents", "parent[]"),
            Property("children", "child[]"),
            Property("nested", "inner[]"),
        ],
        successor_kinds={EdgeTypes.default: ["bla"]},
    )

    return [base, foo, bla, cloud, account, region, parent, child, inner, some_complex]


@fixture
def foo_model(foo_kinds: List[Kind]) -> Model:
    return Model.from_kinds(foo_kinds)


@fixture
def person_model() -> Model:
    zip = StringKind("zip")
    base = ComplexKind(
        "Base",
        [],
        [
            Property("id", "string", required=True, description="Some identifier"),
            Property("kind", "string", required=True, description="Kind if this node."),
            Property("list", "string[]", description="A list of strings."),
            Property("tags", "dictionary[string, string]", description="Key/value pairs."),
            Property("mtime", "datetime", description="Modification time of this node."),
        ],
        metadata={"icon": "icon.svg"},
    )
    address = ComplexKind(
        "Address",
        ["Base"],
        [
            Property("zip", "zip", description="The zip code."),
            Property("city", "string", required=True, description="The name of the city.\nAnd another line."),
        ],
        metadata={"icon": "address.svg"},
    )
    person = ComplexKind(
        "Person",
        ["Base"],
        [
            Property("name", "string", description="The name of the person."),
            Property("address", "Address", description="The address of the person."),
            Property("other_addresses", "dictionary[string, Address]", description="Other addresses."),
            Property("addresses", "Address[]", description="The list of addresses."),
            Property("any", "any", description="Some arbitrary value."),
        ],
    )
    any_foo = ComplexKind(
        "any_foo",
        ["Base"],
        [
            Property("foo", "any", description="Some foo value."),
            Property("test", "string", description="Some test value."),
        ],
    )
    cloud = ComplexKind("cloud", ["Base"], [])
    account = ComplexKind("account", ["Base"], [])
    region = ComplexKind("region", ["Base"], [])
    parent = ComplexKind("parent", ["Base"], [])
    child = ComplexKind("child", ["Base"], [])

    return Model.from_kinds([zip, person, address, base, any_foo, cloud, account, region, parent, child])


@fixture
async def filled_graph_db(graph_db: ArangoGraphDB, foo_model: Model) -> ArangoGraphDB:
    graph_db.db.collection(graph_db.node_history).truncate()
    if await graph_db.db.has_collection("model"):
        graph_db.db.collection("model").truncate()
    await graph_db.wipe()
    await graph_db.merge_graph(create_graph("yes or no"), foo_model)
    return graph_db


@fixture
async def event_graph_db(filled_graph_db: ArangoGraphDB, event_sender: AnalyticsEventSender) -> EventGraphDB:
    return EventGraphDB(filled_graph_db, event_sender)


@fixture
def expander() -> InMemoryTemplateExpander:
    return InMemoryTemplateExpander()


@fixture
async def task_queue() -> AsyncIterator[WorkerTaskQueue]:
    wtq = WorkerTaskQueue()
    await wtq.start()
    yield wtq
    await wtq.stop()


@fixture
def performed_by() -> Dict[str, List[str]]:
    return defaultdict(list)


@fixture
def incoming_tasks() -> List[WorkerTask]:
    return []


@fixture
async def worker(
    task_queue: WorkerTaskQueue, performed_by: Dict[str, List[WorkerId]], incoming_tasks: List[WorkerTask]
) -> AsyncGenerator[Tuple[WorkerTaskDescription, WorkerTaskDescription, WorkerTaskDescription], None]:
    success = WorkerTaskDescription("success_task")
    fail = WorkerTaskDescription("fail_task")
    wait = WorkerTaskDescription("wait_task")
    tag = WorkerTaskDescription(WorkerTaskName.tag)
    validate_config = WorkerTaskDescription(WorkerTaskName.validate_config)

    async def do_work(worker_id: WorkerId, task_descriptions: List[WorkerTaskDescription]) -> None:
        async with task_queue.attach(worker_id, task_descriptions) as tasks:
            while True:
                task: WorkerTask = await tasks.get()
                incoming_tasks.append(task)
                performed_by[task.id].append(worker_id)
                if task.name == success.name:
                    await task_queue.acknowledge_task(worker_id, task.id, {"result": "done!"})
                elif task.name == fail.name:
                    await task_queue.error_task(worker_id, task.id, ";)")
                elif task.name == wait.name:
                    # if we come here, neither success nor failure was given, ignore the task
                    pass
                elif task.name == WorkerTaskName.validate_config:
                    cfg_id = task.attrs["config_id"]
                    if cfg_id == "invalid_config":
                        await task_queue.error_task(worker_id, task.id, "Invalid Config ;)")
                    else:
                        await task_queue.acknowledge_task(worker_id, task.id, None)
                elif task.name == WorkerTaskName.tag:
                    node = task.data["node"]
                    for key in GraphResolver.resolved_ancestors.keys():
                        for section in Section.content:
                            if section in node:
                                node[section].pop(key, None)

                    # update or delete tags
                    if "tags" not in node:
                        node["tags"] = {}

                    if task.data.get("delete"):
                        for a in task.data.get("delete"):  # type: ignore
                            node["tags"].pop(a, None)
                    elif task.data.get("update"):
                        for k, v in task.data.get("update").items():  # type: ignore
                            node["tags"][k] = v

                    # for testing purposes: change revision number
                    kind: str = value_in_path(node, NodePath.reported_kind)  # type: ignore
                    if kind == "bla":
                        node["revision"] = "changed"

                    await task_queue.acknowledge_task(worker_id, task.id, node)
                else:
                    await task_queue.error_task(worker_id, task.id, "Don't know how to handle this task")

    workers = [
        asyncio.create_task(do_work(WorkerId(f"w{a}"), [success, fail, wait, tag, validate_config]))
        for a in range(0, 4)
    ]
    await asyncio.sleep(0)

    yield success, fail, wait
    for worker in workers:
        worker.cancel()


@fixture
def cert_handler() -> Iterator[CertificateHandler]:
    config = empty_config()
    key, certificate = bootstrap_ca()
    temp = TemporaryDirectory()
    yield CertificateHandler(config, key, certificate, Path(temp.name))
    temp.cleanup()


@fixture
def config_handler(task_queue: WorkerTaskQueue, worker: Any, message_bus: MessageBus) -> ConfigHandlerService:
    # Note: the worker fixture is required, since it starts worker tasks
    cfg_db = InMemoryDb(ConfigEntity, lambda c: c.id)
    validation_db = InMemoryDb(ConfigValidation, lambda c: c.id)
    model_db = InMemoryDb(Kind, lambda c: c.fqn)  # type: ignore
    event_sender = InMemoryEventSender()
    core_config = cast(CoreConfig, SimpleNamespace())
    override_service = cast(ConfigOverride, SimpleNamespace(get_override=lambda _: {}, get_all_overrides=lambda: {}))
    return ConfigHandlerService(
        cfg_db, validation_db, model_db, task_queue, message_bus, event_sender, core_config, override_service
    )


@fixture
def core_config_handler_exits() -> List[bool]:
    return []


@fixture
async def core_config_handler(
    message_bus: MessageBus,
    task_queue: WorkerTaskQueue,
    config_handler: ConfigHandler,
    inspector_service: InspectorService,
    core_config_handler_exits: List[bool],
) -> CoreConfigHandler:
    def on_exit() -> None:
        core_config_handler_exits.append(True)

    config = empty_config()
    sender = InMemoryEventSender()
    return CoreConfigHandler(config, message_bus, task_queue, config_handler, sender, inspector_service, on_exit)


@fixture
async def core_config_handler_started(core_config_handler: CoreConfigHandler) -> AsyncIterator[CoreConfigHandler]:
    await core_config_handler.start()
    yield core_config_handler
    await core_config_handler.stop()


@fixture
async def dependencies(
    filled_graph_db: ArangoGraphDB,
    message_bus: MessageBus,
    event_sender: InMemoryEventSender,
    foo_model: Model,
    task_queue: WorkerTaskQueue,
    worker: Tuple[WorkerTaskDescription, WorkerTaskDescription, WorkerTaskDescription],
    expander: TemplateExpander,
    config_handler: ConfigHandler,
    cert_handler: CertificateHandler,
    user_management: UserManagement,
) -> AsyncIterator[Dependencies]:
    db_access = DbAccess(filled_graph_db.db.db, event_sender, NoAdjust(), empty_config())
    model_handler = ModelHandlerStatic(foo_model)
    config = empty_config(["--graphdb-database", "test", "--graphdb-username", "test", "--graphdb-password", "test"])
    deps = Dependencies(
        message_bus=message_bus,
        event_sender=event_sender,
        db_access=db_access,
        model_handler=model_handler,
        worker_task_queue=task_queue,
        config=config,
        template_expander=expander,
        forked_tasks=Queue(),
        config_handler=config_handler,
        cert_handler=cert_handler,
        user_management=user_management,
    )
    await db_access.start()
    yield deps
    await db_access.stop()
    await deps.stop()


@fixture
async def graph_merger(
    foo_model: Model, event_sender: AnalyticsEventSender, default_config: CoreConfig, message_bus: MessageBus
) -> GraphMerger:
    model_handler = ModelHandlerStatic(foo_model)
    return GraphMerger(model_handler, event_sender, default_config, message_bus)


@fixture
async def task_dependencies(graph_merger: GraphMerger, subscription_handler: SubscriptionHandler) -> TaskDependencies:
    return TaskDependencies(graph_merger, subscription_handler.subscribers_by_event)


@fixture
def cli(dependencies: Dependencies) -> CLIService:
    env = {"graph": "ns", "section": "reported"}
    return CLIService(dependencies, all_commands(dependencies), env, alias_names())


@fixture
async def inspector_service(cli: CLIService) -> InspectorService:
    async with InspectorService(cli) as service:
        cli.dependencies.lookup["inspector"] = service
        return service


@fixture
async def package_manager(
    cli: CLIService, config_handler: ConfigHandler, package_entity_db: PackageEntityDb
) -> PackageManager:
    async with PackageManager(package_entity_db, config_handler, lambda at: None, lambda s: None) as service:
        cli.dependencies.lookup["infra_apps_package_manager"] = service
        return service


@fixture
async def graph_manager(
    cli: CLIService, db_access: DbAccess, core_config_handler: CoreConfigHandler, task_handler: TaskHandlerService
) -> GraphManager:
    async with GraphManager(db_access, SnapshotsScheduleConfig({}), core_config_handler, task_handler) as service:
        cli.dependencies.lookup["graph_manager"] = service
        return service


@fixture
async def infra_apps_runtime(cli: CLIService) -> LocalResotocoreAppRuntime:
    runtime = LocalResotocoreAppRuntime(cli)
    cli.dependencies.lookup["infra_apps_runtime"] = runtime
    return runtime


@fixture
async def test_benchmark(config_handler: ConfigHandler) -> Benchmark:
    test_check = ConfigEntity(
        ConfigId(CheckConfigPrefix + "test"),
        {
            "report_check": {
                "provider": "test",
                "service": "test",
                "checks": [
                    {
                        "id": "test_test_some_check",
                        "name": "some_check",
                        "title": "Test",
                        "result_kind": "foo",
                        "categories": [],
                        "risk": "Some risk",
                        "severity": "medium",
                        "detect": {"resoto": "is(foo)"},
                        "remediation": {"text": "Some remediation text", "url": "https://example.com"},
                    }
                ],
            }
        },
    )
    test_benchmark = ConfigEntity(
        ConfigId(BenchmarkConfigPrefix + "test"),
        {
            "report_benchmark": {
                "id": "test",
                "title": "test",
                "framework": "test",
                "clouds": ["test"],
                "version": "1.5",
                "description": "test",
                "children": [
                    {
                        "title": "Section 1",
                        "description": "Test section.",
                        "checks": ["test_test_some_check"],
                    },
                    {
                        "title": "Section 2",
                        "description": "Test section.",
                        "checks": ["test_test_some_check"],
                    },
                ],
                "checks": [],
            }
        },
    )
    await config_handler.put_config(test_check)
    await config_handler.put_config(test_benchmark)
    return BenchmarkConfig.from_config(test_benchmark)


@fixture
async def client_session() -> AsyncIterator[ClientSession]:
    session = ClientSession()
    yield session
    await session.close()


@fixture
def test_workflow() -> Workflow:
    return Workflow(
        TaskDescriptorId("test_workflow"),
        "Speakable name of workflow",
        [
            Step("start", PerformAction("start_collect"), timedelta(seconds=10)),
            Step("act", PerformAction("collect"), timedelta(seconds=10)),
            Step("done", PerformAction("collect_done"), timedelta(seconds=10), StepErrorBehaviour.Stop),
        ],
        [EventTrigger("start me up"), TimeTrigger("1 1 1 1 1")],
    )


@fixture
def additional_workflows() -> List[Workflow]:
    return [
        Workflow(
            TaskDescriptorId("sleep_workflow"),
            "Speakable name of workflow",
            [Step("sleep", ExecuteCommand("sleep 0.1"), timedelta(seconds=10))],
            triggers=[],
            on_surpass=TaskSurpassBehaviour.Wait,
        ),
        Workflow(
            TaskDescriptorId("wait_for_collect_done"),
            "Wait for collect",
            [
                Step("wait", WaitForEvent("collected", {}), timedelta(seconds=10)),
                Step("wait_for_collect_done", WaitForCollectDone(), timedelta(seconds=10)),
            ],
            triggers=[],
            on_surpass=TaskSurpassBehaviour.Wait,
        ),
    ]


@fixture
def test_wait_workflow() -> Workflow:
    return Workflow(
        TaskDescriptorId("test_workflow"),
        "Speakable name of workflow",
        [
            Step("start", PerformAction("start_collect"), timedelta(seconds=10)),
            Step("wait", WaitForEvent("godot", {"a": 1}), timedelta(seconds=10)),
            Step("emit_event", EmitEvent(Event("hello", {"a": 1})), timedelta(seconds=10)),
            Step("collect", PerformAction("collect"), timedelta(seconds=10)),
            Step("done", PerformAction("collect_done"), timedelta(seconds=10), StepErrorBehaviour.Stop),
        ],
        [EventTrigger("start me up")],
    )


@fixture
def workflow_instance(
    task_dependencies: TaskDependencies,
    test_wait_workflow: Workflow,
) -> Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]]:
    td = timedelta(seconds=100)
    sub1 = Subscription("start_collect", True, td)
    sub2 = Subscription("collect", True, td)
    sub3 = Subscription("collect_done", True, td)
    s1 = Subscriber.from_list(SubscriberId("s1"), [sub1, sub2, sub3])
    s2 = Subscriber.from_list(SubscriberId("s2"), [sub2, sub3])
    subscriptions = {"start_collect": [s1], "collect": [s1, s2], "collect_done": [s1, s2]}
    w, _ = RunningTask.empty(test_wait_workflow, evolve(task_dependencies, subscribers_by_event=lambda: subscriptions))
    w.received_messages = [
        Action("start_collect", w.id, "start"),
        ActionDone("start_collect", w.id, "start", s1.id),
        ActionDone("start_collect", w.id, "start", s2.id),
        Event("godot", {"a": 1, "b": 2}),
        Action("collect", w.id, "collect"),
        ActionDone("collect", w.id, "collect", s1.id),
    ]
    w.move_to_next_state()
    return w, s1, s2, subscriptions


@fixture
async def subscription_handler(message_bus: MessageBus) -> AsyncIterator[SubscriptionHandler]:
    in_mem = InMemoryDb(Subscriber, lambda x: x.id)
    async with SubscriptionHandler(in_mem, message_bus) as handler:
        yield handler


@fixture
async def task_handler(
    running_task_db: RunningTaskDb,
    job_db: JobDb,
    message_bus: MessageBus,
    event_sender: AnalyticsEventSender,
    subscription_handler: SubscriptionHandler,
    graph_merger: GraphMerger,
    cli: CLIService,
    test_workflow: Workflow,
    additional_workflows: List[Workflow],
) -> AsyncGenerator[TaskHandlerService, None]:
    config = empty_config()
    task_handler = TaskHandlerService(
        running_task_db, job_db, message_bus, event_sender, subscription_handler, graph_merger, Scheduler(), cli, config
    )
    task_handler.task_descriptions = additional_workflows + [test_workflow]
    cli.dependencies.lookup["task_handler"] = task_handler
    async with task_handler:
        yield task_handler


@fixture
def tmp_directory() -> Generator[str, None, None]:
    tmp_dir: Optional[str] = None
    try:
        tmp_dir = tempfile.mkdtemp()
        yield tmp_dir
    finally:
        if tmp_dir:
            shutil.rmtree(tmp_dir)


@fixture
async def echo_http_server() -> AsyncIterator[Tuple[int, List[Tuple[Request, Json]]]]:
    requests = []

    async def add_request(request: Request) -> Response:
        requests.append((request, await request.json()))
        status = 500 if request.path.startswith("/fail") else 200
        return Response(status=status)

    app = Application()
    app.add_routes([route(METH_ANY, "/{tail:.+}", add_request)])
    server = TestServer(app)
    await server.start_server()
    yield server.port, requests  # type: ignore
    await server.close()


@fixture()
async def merge_handler(
    message_bus: MessageBus,
    subscription_handler: SubscriptionHandler,
    task_handler: TaskHandlerService,
    db_access: DbAccess,
    foo_model: Model,
) -> AsyncGenerator[MergeOuterEdgesHandler, None]:
    model_handler = ModelHandlerStatic(foo_model)
    handler = MergeOuterEdgesHandler(message_bus, subscription_handler, task_handler, db_access, model_handler)
    await handler.start()
    yield handler
    await handler.stop()


@fixture
def console_renderer() -> ConsoleRenderer:
    tty_columns, tty_rows = shutil.get_terminal_size(fallback=(80, 25))
    return ConsoleRenderer(tty_columns, tty_rows, ConsoleColorSystem.from_name(Console().color_system or "standard"))


@fixture
async def user_management(
    db_access: DbAccess, config_handler: ConfigHandler, event_sender: AnalyticsEventSender
) -> UserManagementService:
    return UserManagementService(db_access, config_handler, event_sender)


@fixture
async def system_data_db(test_db: StandardDatabase) -> AsyncIterator[SystemDataDb]:
    with suppress(Exception):
        system = SystemData(uuid_str(), utc(), 1)
        test_db.insert_document("system_data", {"_key": "system", **to_js(system)}, overwrite=False)
    with suppress(Exception):
        test_db.insert_document(
            "system_data", {"_key": "ca", "key": "private_key", "certificate": "some cert"}, overwrite=False
        )
    async_db = AsyncArangoDB(test_db)
    yield SystemDataDb(async_db)
    test_db.collection("system_data").delete({"_key": "ca"})


async def eventually(
    predicate: Callable[[], bool],
    timeout: timedelta = timedelta(seconds=5),
    interval: timedelta = timedelta(seconds=0.1),
) -> None:
    async def wait_for_condition() -> None:
        while not predicate():
            await asyncio.sleep(interval.total_seconds())

    await asyncio.wait_for(wait_for_condition(), timeout.total_seconds())



from pytest import fixture

from resotocore.cli.cli import CLIService
from resotocore.config import ConfigEntity
from resotocore.ids import ConfigId
from resotocore.report import BenchmarkConfigRoot, CheckConfigRoot
from resotocore.report.inspector_service import InspectorService, check_id, benchmark_id
from resotocore.report.report_config import (
    config_model,
    ReportCheckCollectionConfig,
    BenchmarkConfig,
)
from resotocore.types import Json
from resotocore.util import partition_by


@fixture
async def inspector_service_with_test_benchmark(
    cli: CLIService, inspection_check_collection: Json, benchmark: Json
) -> InspectorService:
    service = InspectorService(cli)
    await service.config_handler.put_config(ConfigEntity(check_id("test"), inspection_check_collection))
    await service.config_handler.put_config(ConfigEntity(benchmark_id("test"), benchmark))
    return service


@fixture
def inspection_check_collection() -> Json:
    return dict(
        report_check=dict(
            provider="test",
            service="test",
            checks=[
                dict(
                    title="search",
                    name="search",
                    result_kind="foo",
                    categories=["test"],
                    severity="critical",
                    risk="",
                    # we use a query with a template here
                    detect={"resoto": "is({{foo_kind}})"},
                    default_values={"foo_kind": "foo"},
                    remediation=dict(text="", url="", action={}),
                ),
                dict(
                    title="cmd",
                    name="cmd",
                    result_kind="foo",
                    categories=["test"],
                    severity="critical",
                    detect={"resoto_cmd": "search is(foo) | jq --no-rewrite ."},
                    risk="",
                    remediation=dict(text="", url="", action={}),
                ),
            ],
        )
    )


@fixture
def benchmark() -> Json:
    return dict(
        report_benchmark=dict(
            title="test_benchmark",
            description="test_benchmark",
            id="test_benchmark",
            framework="test",
            version="1.0",
            checks=["test_test_search", "test_test_cmd"],
            clouds=["test"],
        )
    )


async def test_config_model() -> None:
    models = config_model()
    assert len(models) == 6


async def test_list_inspect_checks(inspector_service: InspectorService) -> None:
    # list all available checks
    all_checks = {i.id: i for i in await inspector_service.list_checks()}
    assert len(all_checks) >= 30

    # use different filter options. The more filter are used, fewer results are returned
    filter_options = dict(
        provider="aws",
        service="ec2",
        category="security",
        kind="aws_ec2_instance",
        check_ids=["aws_ec2_internet_facing_with_instance_profile"],
    )
    last_len = len(all_checks)
    for options in range(1, len(filter_options)):
        args = dict(list(filter_options.items())[0:options])
        matching_checks = [i for i in await inspector_service.list_checks(**args)]  # type: ignore
        assert len(matching_checks) > 0
        assert len(matching_checks) <= last_len
        last_len = len(matching_checks)
    assert last_len < len(all_checks)


async def test_perform_benchmark(inspector_service_with_test_benchmark: InspectorService) -> None:
    inspector = inspector_service_with_test_benchmark
    result = await inspector.perform_benchmark(inspector.cli.env["graph"], "test")
    assert result.checks[0].number_of_resources_failing == 11
    assert result.checks[1].number_of_resources_failing == 11
    filtered = result.filter_result(filter_failed=True)
    assert filtered.checks[0].number_of_resources_failing == 11
    assert filtered.checks[1].number_of_resources_failing == 11
    passing, failing = result.passing_failing_checks_for_account("sub_root")
    assert len(passing) == 0
    assert len(failing) == 2
    passing, failing = result.passing_failing_checks_for_account("does_not_exist")
    assert len(passing) == 2
    assert len(failing) == 0


async def test_benchmark_node_result(inspector_service_with_test_benchmark: InspectorService) -> None:
    inspector = inspector_service_with_test_benchmark
    result = await inspector.perform_benchmark(inspector.cli.env["graph"], "test")
    node_edge_list = result.to_graph()
    nodes, edges = partition_by(lambda x: x["type"] == "node", node_edge_list)
    assert len(node_edge_list) == 5  # 3 nodes + 2 edges
    assert len(nodes) == 3
    assert len(edges) == 2
    for edge in edges:
        assert edge["from"] == result.node_id


async def test_predefined_checks(inspector_service: InspectorService) -> None:
    checks = ReportCheckCollectionConfig.from_files()
    assert len(checks) > 0
    for name, check in checks.items():
        assert (await inspector_service.validate_check_collection_config({CheckConfigRoot: check})) is None


async def test_predefined_benchmarks(inspector_service: InspectorService) -> None:
    benchmarks = BenchmarkConfig.from_files()
    assert len(benchmarks) > 0
    for name, check in benchmarks.items():
        config = {BenchmarkConfigRoot: check}
        assert (await inspector_service.validate_benchmark_config(config)) is None
        benchmark = BenchmarkConfig.from_config(ConfigEntity(ConfigId("test"), config))
        assert benchmark.clouds == ["aws"]


async def test_list_failing(inspector_service_with_test_benchmark: InspectorService) -> None:
    inspector = inspector_service_with_test_benchmark
    graph = inspector.cli.env["graph"]
    search_res = [r async for r in await inspector.list_failing_resources(graph, "test_test_search")]
    assert len(search_res) == 11
    cmd_res = [r async for r in await inspector.list_failing_resources(graph, "test_test_cmd")]
    assert len(cmd_res) == 11
    search_res_account = [r async for r in await inspector.list_failing_resources(graph, "test_test_search", ["n/a"])]
    assert len(search_res_account) == 0
    cmd_res_account = [r async for r in await inspector.list_failing_resources(graph, "test_test_cmd", ["n/a"])]
    assert len(cmd_res_account) == 0



from resotocore.config import ConfigEntity, ConfigValidation
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EntityDb, EventEntityDb, ArangoEntityDb
from resotocore.ids import ConfigId


# Database to store config entities
ConfigEntityDb = EntityDb[ConfigId, ConfigEntity]
EventConfigEntityDb = EventEntityDb[ConfigId, ConfigEntity]


def config_entity_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[ConfigId, ConfigEntity]:
    return ArangoEntityDb(db, collection, ConfigEntity, lambda k: k.id)


# Database to store config entity models
ConfigValidationEntityDb = EntityDb[str, ConfigValidation]
EventConfigValidationEntityDb = EventEntityDb[str, ConfigValidation]


def config_validation_entity_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[str, ConfigValidation]:
    return ArangoEntityDb(db, collection, ConfigValidation, lambda k: k.id)
